import uuid
from functools import lru_cache
from typing import Union

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction
from django.utils.translation import gettext_lazy as _
from rest_framework.serializers import ValidationError

from .service_identifier import service_id


@lru_cache(maxsize=None)
def resource_type_cache(content_type_id):
    return ContentType.objects.get_for_id(content_type_id).resource_type


class UnamangedResourceException(Exception):
    pass


class ResourceType(models.Model):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from ansible_base.resource_registry.registry import get_registry

        self.resource_registry = get_registry()

    content_type = models.OneToOneField(
        ContentType, on_delete=models.CASCADE, related_name="resource_type", unique=True, help_text=_("The content type for this resource type.")
    )
    externally_managed = models.BooleanField(help_text=_("Is this resource type managed externally from this service."))
    name = models.CharField(max_length=256, unique=True, db_index=True, editable=False, blank=False, null=False, help_text=_("The name of this resource type."))

    @property
    def serializer_class(self):

        return self.get_resource_config().managed_serializer

    @property
    def can_be_managed(self):
        return self.serializer_class is not None

    def get_resource_config(self):
        return self.resource_registry.get_config_for_model(model=ContentType.objects.get_for_id(self.content_type_id).model_class())

    def get_conflicting_resource(self, resource_data):
        qfilter = {}

        if not self.can_be_managed:
            raise UnamangedResourceException(f"Resource type {self.name} does not have a managed serializer.")

        serializer = self.serializer_class(data=resource_data)
        serializer.is_valid(raise_exception=True)
        serialized_data = serializer.validated_data

        for field in self.serializer_class.UNIQUE_FIELDS:
            qfilter[field] = serialized_data[field]

        try:
            return Resource.get_resource_for_object(self.content_type.get_object_for_this_type(**qfilter))
        except ObjectDoesNotExist:
            return None


class Resource(models.Model):
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, related_name="resources", help_text=_("The content type for this resource."))

    # this has to accommodate integer and UUID object IDs
    object_id = models.TextField(null=False, db_index=True, help_text=_("The object id for this resource."))
    content_object = GenericForeignKey('content_type', 'object_id')

    service_id = models.UUIDField(
        null=False,
        default=service_id,
        help_text=_("ID of the service responsible for managing this resource."),
    )

    # we're not using this as the primary key because the ansible_id can change if the object is
    # externally managed.
    ansible_id = models.UUIDField(default=uuid.uuid4, db_index=True, unique=True, help_text=_("A unique ID identifying this resource by the resource server."))

    # human readable name for the resource
    name = models.CharField(max_length=512, null=True, help_text=_("The name of this resource."))

    is_partially_migrated = models.BooleanField(
        default=False,
        help_text=_("A flag indicating that the resource has been copied into the resource server, but the service_id hasn't been updated yet."),
    )

    def summary_fields(self):
        return {"ansible_id": self.ansible_id, "resource_type": self.resource_type}

    @property
    def resource_type(self):
        return resource_type_cache(self.content_type.pk).name

    @property
    def resource_type_obj(self):
        return resource_type_cache(self.content_type.pk)

    class Meta:
        unique_together = ('content_type', 'object_id')
        indexes = [
            models.Index(fields=["content_type", "object_id"]),
        ]

    def update_from_content_object(self):
        """
        Update any cached attributes from the Resource's content_object
        """
        name_field = self.content_type.resource_type.get_resource_config().name_field

        if hasattr(self.content_object, name_field):
            name = getattr(self.content_object, name_field)[:512]
            if self.name != name:
                self.name = name
                self.save()

    @classmethod
    def get_resource_for_object(cls, obj):
        """
        Get the Resource instances for another model instance.
        """
        return cls.objects.get(object_id=obj.pk, content_type=ContentType.objects.get_for_model(obj).pk)

    def delete_resource(self):
        from ..signals.handlers import no_reverse_sync

        if not self.content_type.resource_type.can_be_managed:
            raise ValidationError({"resource_type": _(f"Resource type: {self.content_type.resource_type.name} cannot be managed by Resources.")})

        with transaction.atomic():
            with no_reverse_sync():
                self.content_object.delete()
            self.delete()

    @classmethod
    def create_resource(
        cls,
        resource_type: ResourceType,
        resource_data: dict,
        ansible_id: Union[str, uuid.UUID, None] = None,
        service_id: Union[str, uuid.UUID, None] = None,
        is_partially_migrated=False,
    ):
        from ..signals.handlers import no_reverse_sync

        c_type = resource_type.content_type
        serializer = resource_type.serializer_class(data=resource_data)
        serializer.is_valid(raise_exception=True)
        resource_data = serializer.validated_data
        processor = serializer.get_processor()

        with transaction.atomic():
            ObjModel = c_type.model_class()
            content_object = processor(ObjModel())
            with no_reverse_sync():
                content_object.save(resource_data, is_new=True)
            resource = cls.objects.get(object_id=content_object.instance.pk, content_type=c_type)
            resource.is_partially_migrated = is_partially_migrated

            if ansible_id:
                resource.ansible_id = ansible_id
            if service_id:
                resource.service_id = service_id
            resource.save()

            return resource

    def update_resource(self, resource_data: dict, ansible_id=None, is_partially_migrated=None, partial=False, service_id: Union[str, uuid.UUID, None] = None):
        from ..signals.handlers import no_reverse_sync

        resource_type = self.content_type.resource_type

        serializer = resource_type.serializer_class(data=resource_data, partial=partial)
        serializer.is_valid(raise_exception=True)
        resource_data = serializer.validated_data

        processor = serializer.get_processor()

        with transaction.atomic():
            if ansible_id:
                self.ansible_id = ansible_id
            if service_id:
                self.service_id = service_id
            if is_partially_migrated is not None:
                self.is_partially_migrated = is_partially_migrated
            self.save()

            content_object = processor(self.content_object)
            with no_reverse_sync():
                content_object.save(resource_data)


# This is a separate function so that it can work with models from apps in the
# post migration signal.
def init_resource_from_object(obj, resource_model=None, resource_type=None, resource_config=None):
    """
    Initialize a new Resource object from another model instance.
    """
    if resource_type is None:
        c_type = ContentType.objects.get_for_model(obj)
        resource_type = c_type.resource_type
        assert resource_type is not None

    if resource_config is None:
        resource_config = resource_type.get_resource_config()

    if resource_model is None:
        resource_model = Resource

    resource = resource_model(object_id=obj.pk, content_type=resource_type.content_type)
    if hasattr(obj, resource_config.name_field):
        resource.name = str(getattr(obj, resource_config.name_field))[:512]

    return resource



import hashlib
import json
from typing import Callable, Optional

from django.utils.translation import gettext_lazy as _
from rest_framework import serializers

from ansible_base.resource_registry.models import Resource
from ansible_base.resource_registry.registry import get_registry


class AnsibleResourceManyRelated(serializers.Field):
    def __init__(self, resource_type, **kwargs):
        self.resource_type = resource_type
        super().__init__(**kwargs)

    def to_representation(self, value):
        pks = value
        if not isinstance(value, list):
            pks = value.all().values_list("pk", flat=True)

        pks = [str(x) for x in pks]
        return Resource.objects.filter(content_type__resource_type__name=self.resource_type, object_id__in=pks).values_list("ansible_id", flat=True)


class AnsibleResourceForeignKeyField(serializers.UUIDField):
    default_error_messages = {
        'invalid': _('Must be a valid UUID.'),
        'does_not_exist': _('Invalid ansible id "{ansible_id}" - resource does not exist.'),
    }

    def __init__(self, resource_type, **kwargs):
        self.resource_type = resource_type
        super().__init__(**kwargs)

    # Convert ansible ID to internal object ID
    def to_internal_value(self, data):
        ansible_id = super().to_internal_value(data)

        try:
            resource = Resource.objects.get(content_type__resource_type__name=self.resource_type, ansible_id=ansible_id)
        except Resource.DoesNotExist:
            self.fail('does_not_exist', ansible_id=data)

        return resource.content_object

    def get_attribute(self, instance):
        # If the model doesn't have an attribute with the given field name, return None. This is
        # mostly here to keep Hub from breaking, which doesn't have organizations yet.
        obj = getattr(instance, self.field_name, None)
        if obj is None:
            return None
        resource = Resource.objects.get(content_type__resource_type__name=self.resource_type, object_id=obj.pk)

        return resource.ansible_id


class SharedResourceTypeSerializer(serializers.Serializer):
    """
    This is the base class for resource type serializers. Serializers that extend this class are
    intended to represent data that is common across all instances of a service. Fields on serializers
    that implement this base class are required to be readable and writeable across all services.
    Fields should only serialize concrete data on the resource model itself and must not include
    many to many or many to one relations. Fields declared here will be used to compute hashes
    for keeping resources in sync between services.
    """

    def __init__(self, instance=None, **kwargs):
        if instance:
            processor = self.get_processor()
            instance = processor(instance).pre_serialize()

        super().__init__(instance, **kwargs)

    # Required. This field defines the shared type and will include `shared.` prefix in the
    # ResourceType definition model. This field must be unique and cannot be changed.
    RESOURCE_TYPE = None

    # Optional. This indicates a serializer to use to display additional data about a resource.
    # Data defined in this serializer can contain many to one, many to many and computed fields.
    # Fields here may be writeable, but are not expected to be. Data from this serializer wil
    # not be used to compute hashes for sync.
    ADDITIONAL_DATA_SERIALIZER = None

    def get_hash(self, field: Optional[str] = None, hasher: Callable = hashlib.sha256):
        """
        Takes an instance, serialize it and take the .data or the specified field
        as input for the hasher function.
        """
        serialized_data = self.data
        if field:
            serialized_data = serialized_data[field]
        metadata_json = json.dumps(serialized_data, sort_keys=True).encode("utf-8")
        return hasher(metadata_json).hexdigest()

    @classmethod
    def get_processor(cls):
        return get_registry().api_config.get_processor(f"shared.{cls.RESOURCE_TYPE}")

    @classmethod
    def get_additional_data(cls, instance):
        if cls.ADDITIONAL_DATA_SERIALIZER is None:
            return None

        processor = cls.get_processor()
        instance = processor(instance).pre_serialize_additional()

        return cls.ADDITIONAL_DATA_SERIALIZER(instance)



# Generated by Django 4.2.16 on 2024-11-21 11:14

import ansible_base.authentication.models.authenticator
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dab_authentication', '0014_authenticator_auto_migrate_users_to'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authenticator',
            name='category',
            field=models.CharField(default=None, help_text='The base type of this authenticator.', max_length=30),
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='configuration',
            field=models.JSONField(blank=True, default=dict, help_text='The required configuration for this source.'),
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='create_objects',
            field=models.BooleanField(default=True, help_text='Allow authenticator to create objects (users, teams, organizations).'),
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='enabled',
            field=models.BooleanField(default=False, help_text='Should this authenticator be enabled.'),
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='modified_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='name',
            field=models.CharField(help_text='The name of this resource.', max_length=512, unique=True),
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='order',
            field=models.IntegerField(default=ansible_base.authentication.models.authenticator.get_next_authenticator_order, help_text='The order in which an authenticator will be tried. This only pertains to username/password authenticators.'),
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='remove_users',
            field=models.BooleanField(default=True, help_text='When a user authenticates from this source should they be removed from any other groups they were previously added to.'),
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='slug',
            field=models.SlugField(default=None, editable=False, help_text='An immutable identifier for the authenticator.', max_length=1024, unique=True),
        ),
        migrations.AlterField(
            model_name='authenticator',
            name='type',
            field=models.CharField(editable=False, help_text='The type of authentication service this is.', max_length=256),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='authenticator',
            field=models.ForeignKey(help_text='The authenticator this mapping belongs to.', on_delete=django.db.models.deletion.CASCADE, related_name='authenticator_maps', to='dab_authentication.authenticator'),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='map_type',
            field=models.CharField(choices=[('allow', 'allow'), ('is_superuser', 'is_superuser'), ('role', 'role'), ('organization', 'organization'), ('team', 'team')], default='team', help_text='What will the map grant the user? System access (allow) a team or organization membership, the superuser flag or a role in the system.', max_length=17),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='modified_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='name',
            field=models.CharField(help_text='The name of this resource.', max_length=512),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='order',
            field=models.PositiveIntegerField(default=0, help_text='The order in which this rule should be processed, smaller numbers are of higher precedence. Items with the same order will be executed in random order.'),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='organization',
            field=models.CharField(blank=True, default=None, help_text='An organization name this rule works on.', max_length=512, null=True),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='revoke',
            field=models.BooleanField(default=False, help_text='Should we revoke the permission if a user does not meet this rule.'),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='role',
            field=models.CharField(blank=True, default=None, help_text='The role this map will grant the authenticating user to the targeted object.', max_length=512, null=True),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='team',
            field=models.CharField(blank=True, default=None, help_text='A team name this rule works on.', max_length=512, null=True),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='triggers',
            field=models.JSONField(blank=True, default=dict, help_text='Trigger information for this rule.'),
        ),
        migrations.AlterField(
            model_name='authenticatormap',
            name='revoke',
            field=models.BooleanField(default=False, help_text='Revoke the permission if a user does not meet this rule.'),
        ),
        migrations.AddField(
            model_name='authenticatormap',
            name='enabled',
            field=models.BooleanField(default=True, help_text='Enables or disables this authenticator map'),
        ),
    ]



import logging

from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from rest_framework.utils.urls import remove_query_param, replace_query_param

from ansible_base.lib.utils.settings import get_setting

DEFAULT_PAGE_SIZE = 50
DEFAULT_MAX_PAGE_SIZE = 200

logger = logging.getLogger('ansible_base.rest_pagination.default_paginator')


class DefaultPaginator(PageNumberPagination):
    page_size_query_param = 'page_size'

    def __init__(self, *args, **kwargs):
        # Since this class is loaded during settings load we have to get these settings on init instead of as class variables
        self.page_size = get_setting('DEFAULT_PAGE_SIZE', DEFAULT_PAGE_SIZE)
        self.max_page_size = get_setting('MAX_PAGE_SIZE', DEFAULT_MAX_PAGE_SIZE)

    # The default PageNumberPagination wants to deliver an absolute URL instead of a relative one so we override these functions to default to relative URLs
    def get_next_link(self):
        if not self.page.has_next():
            return None
        url = self.request.get_full_path()
        url = url.encode('utf-8')
        page_number = self.page.next_page_number()
        return replace_query_param(url, self.page_query_param, page_number)

    def get_previous_link(self):
        if not self.page.has_previous():
            return None
        url = self.request.get_full_path()
        url = url.encode('utf-8')
        page_number = self.page.previous_page_number()
        if page_number == 1:
            return remove_query_param(url, self.page_query_param)
        return replace_query_param(url, self.page_query_param, page_number)

    # We need something to capture the count_disabled parameter from the request
    def paginate_queryset(self, queryset, request, view=None):
        self.count_disabled = 'count_disabled' in request.query_params
        # AWX had optimizations for larger models in here we are starting without
        # https://github.com/ansible/awx/commit/c0d9600b66a8b271e4da9b74d0706c6d40ee62bf#diff-d87f638f22a7a23098410cc35ff29b4d811332b31122a04e8cb4f4d00143aa62R57
        return super().paginate_queryset(queryset, request, view)

    # If the count_disabled was set, do not return the count otherwise take the normal action
    def get_paginated_response(self, data):
        if self.count_disabled:
            return Response({'results': data})

        return super().get_paginated_response(data)



from django.conf import settings
from django.utils.translation import gettext_lazy as _
from rest_framework.viewsets import ReadOnlyModelViewSet

from ansible_base.activitystream.filtering import ActivityStreamFilterBackend
from ansible_base.activitystream.models import Entry
from ansible_base.activitystream.serializers import EntrySerializer
from ansible_base.lib.utils.views.django_app_api import AnsibleBaseDjangoAppApiView
from ansible_base.lib.utils.views.permissions import IsSuperuser


def calculate_filter_backends():
    """
    Calculate the filter backends for the Activity Stream.

    We want to use the ActivityStreamFilterBackend for the FieldLookupBackend, but we want to keep the
    default filter backends (in order) for everything else.
    """
    default_filter_backends = AnsibleBaseDjangoAppApiView.filter_backends
    filter_backends = []
    added_activity_stream_backend = False
    for backend in default_filter_backends:
        if backend.__name__ == 'FieldLookupBackend':
            filter_backends.append(ActivityStreamFilterBackend)
            added_activity_stream_backend = True
        else:
            filter_backends.append(backend)

    if not added_activity_stream_backend:
        filter_backends.append(ActivityStreamFilterBackend)

    return filter_backends


class EntryReadOnlyViewSet(ReadOnlyModelViewSet, AnsibleBaseDjangoAppApiView):
    """
    API endpoint that allows for read-only access to activity stream entries.
    """

    queryset = Entry.objects.order_by('-id')
    serializer_class = EntrySerializer
    filter_backends = calculate_filter_backends()

    def get_permissions(self):
        """
        If the RBAC app is enabled, then we delegate to it to manage permissions.
        Otherwise, we require superuser status.
        """
        if 'ansible_base.rbac' in settings.INSTALLED_APPS:
            # TODO: permission_classes = [AnsibleBaseObjectPermissions] is not applicable now,
            #  set to default permission class
            return super().get_permissions()
        else:
            permission_classes = [IsSuperuser]
        return [permission() for permission in permission_classes]

    def get_view_name(self):
        return _('Activity Stream Entries')



# Generated by Django 4.2.16 on 2024-11-21 11:14

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dab_activitystream', '0003_alter_entry_options'),
    ]

    operations = [
        migrations.AlterField(
            model_name='entry',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created.'),
        ),
        migrations.AlterField(
            model_name='entry',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
    ]



import logging

from django.utils.translation import gettext_lazy as _
from social_core.backends.keycloak import KeycloakOAuth2

from ansible_base.authentication.authenticator_plugins.base import AbstractAuthenticatorPlugin, BaseAuthenticatorConfiguration
from ansible_base.authentication.social_auth import SocialAuthMixin
from ansible_base.lib.serializers.fields import CharField, URLField

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.keycloak')


class KeycloakConfiguration(BaseAuthenticatorConfiguration):
    documentation_url = "https://python-social-auth.readthedocs.io/en/latest/backends/keycloak.html"

    ACCESS_TOKEN_URL = URLField(
        help_text=_("Location where this app can fetch the user's token from."),
        default="https://keycloak.example.com/auth/realms/<my_realm>/protocol/openid-connect/token",
        allow_null=False,
        ui_field_label=_('Keycloak Access Token URL'),
    )
    AUTHORIZATION_URL = URLField(
        help_text=_("Location to redirect the user to during the login flow."),
        default="https://keycloak.example.com/auth/realms/<my_realm>/protocol/openid-connect/auth",
        allow_null=False,
        ui_field_label=_('Keycloak Provider URL'),
    )
    KEY = CharField(
        help_text=_("The OIDC key (Client ID) from your Keycloak installation."),
        allow_null=False,
        ui_field_label=_('Keycloak OIDC Key'),
    )
    PUBLIC_KEY = CharField(
        help_text=_("RS256 public key provided by your Keycloak realm."),
        allow_null=False,
        ui_field_label=_('Keycloak Public Key'),
    )
    SECRET = CharField(
        help_text=_("The OIDC secret (Client Secret) from your Keycloak installation."),
        allow_null=True,
        ui_field_label=_('Keycloak OIDC Secret'),
    )

    GROUPS_CLAIM = CharField(
        help_text=_("The JSON key used to extract the user's groups from the ID token or userinfo endpoint."),
        required=False,
        allow_null=True,
        default="Group",
        ui_field_label=_("Groups Claim"),
    )


class AuthenticatorPlugin(SocialAuthMixin, KeycloakOAuth2, AbstractAuthenticatorPlugin):
    configuration_class = KeycloakConfiguration
    type = "keycloak"
    logger = logger
    category = "sso"
    configuration_encrypted_fields = ['SECRET']

    @property
    def groups_claim(self):
        return self.setting('GROUPS_CLAIM')

    def extra_data(self, user, backend, response, *args, **kwargs):
        for perm in ["is_superuser", "is_platform_auditor"]:
            if perm in response:
                kwargs["social"].extra_data[perm] = response[perm]
        data = super().extra_data(user, backend, response, *args, **kwargs)
        return data

    def get_user_groups(self, extra_groups=[]):
        return extra_groups

    def get_alternative_uid(self, **kwargs):
        return kwargs.get("response", {}).get("sub", None)



import logging
from collections import OrderedDict
from functools import lru_cache

from django.contrib.auth.backends import ModelBackend

from ansible_base.authentication.authenticator_plugins.utils import get_authenticator_plugin
from ansible_base.authentication.models import Authenticator

logger = logging.getLogger('ansible_base.authentication.backend')


@lru_cache(maxsize=1)
def get_authentication_backends(last_updated):
    # last_updated is primarily here as a cache busting mechanism
    authentication_backends = OrderedDict()

    for database_authenticator in Authenticator.objects.filter(enabled=True).order_by('order'):
        try:
            authentication_backends[database_authenticator.id] = get_authenticator_plugin(database_authenticator.type)
        except ImportError:
            continue
        authenticator_object = authentication_backends[database_authenticator.id]
        authenticator_object.update_if_needed(database_authenticator)
    return authentication_backends


class AnsibleBaseAuth(ModelBackend):
    def authenticate(self, request, *args, **kwargs):
        from ansible_base.authentication.social_auth import SOCIAL_AUTH_PIPELINE_FAILED_STATUS

        logger.debug("Starting AnsibleBaseAuth authentication")

        # Query the database for the most recently last modified timestamp.
        # This will be used as a cache key for the cached function get_authentication_backends below
        last_modified_item = Authenticator.objects.values("modified").order_by("-modified").first()
        last_modified = None if last_modified_item is None else last_modified_item.get('modified')

        for authenticator_id, authenticator_object in get_authentication_backends(last_modified).items():
            try:
                user = authenticator_object.authenticate(request, *args, **kwargs)
            except Exception:
                logger.exception(f"Exception raised while trying to authenticate with {authenticator_object.database_instance.name}")
                continue

            # Social Auth pipeline can return status string when update_user_claims fails (authentication maps deny access)
            if user == SOCIAL_AUTH_PIPELINE_FAILED_STATUS:
                continue

            if user:
                # The local authenticator handles this but we want to check this for other authentication types
                if not getattr(user, 'is_active', True):
                    logger.warning(
                        f'User {user.username} attempted to login from authenticator with ID "{authenticator_id}" their user is inactive, denying permission'
                    )
                    return None

                logger.info(f'User {user.username} logged in from authenticator with ID "{authenticator_id}"')
                return user
        return None



from collections import namedtuple
from typing import List, Optional

from django.contrib.auth import authenticate
from django.utils.translation import gettext_lazy as _

from ansible_base.resource_registry.utils.resource_type_processor import ResourceTypeProcessor

ParentResource = namedtuple("ParentResource", ["model", "field_name"])
SharedResource = namedtuple("SharedResource", ["serializer", "is_provider"])


def get_concrete_model(model):
    _model = model
    while _model._meta.proxy_for_model:
        _model = _model._meta.proxy_for_model

    return _model


class ServiceAPIConfig:
    """
    This will be the interface for configuring the resource registry for each service.
    """

    _default_resource_processors = {
        "shared.team": ResourceTypeProcessor,
        "shared.organization": ResourceTypeProcessor,
        "shared.user": ResourceTypeProcessor,
    }

    custom_resource_processors = {}

    service_type = None

    @staticmethod
    def authenticate_local_user(username: str, password: str):
        """
        Return User instance or None
        """
        return authenticate(username=username, password=password)

    @classmethod
    def get_processor(cls, resource_type):
        combined_processors = {**cls._default_resource_processors, **cls.custom_resource_processors}
        return combined_processors[resource_type]


class ResourceConfig:
    model_label = None
    model = None
    externally_managed = None
    managed_serializer = None
    parent_resources = None
    actions = None
    name_field = None

    def __init__(
        self, model, shared_resource: Optional[SharedResource] = None, parent_resources: Optional[List[ParentResource]] = None, name_field: Optional[str] = None
    ):
        model = get_concrete_model(model)
        self.model_label = model._meta.label

        managed_serializer = None
        externally_managed = False
        if name_field is None:
            name_field = "name"

        if shared_resource:
            managed_serializer = shared_resource.serializer
            externally_managed = not shared_resource.is_provider

        parent_map = {}
        if parent_resources:
            for parent in parent_resources:
                parent_map[parent.model._meta.label] = parent

        self.model = model
        self.externally_managed = externally_managed
        self.managed_serializer = managed_serializer
        self.parent_resources = parent_map
        self.name_field = name_field


class ResourceRegistry:
    registry = {}

    def __init__(self, resource_list: List[ResourceConfig], service_api_config: ServiceAPIConfig = None):
        self.api_config = service_api_config
        for r in resource_list:
            self.registry[r.model_label] = r

    def get_resources(self):
        return self.registry

    def get_config_for_model(self, model=None, model_label=None) -> ResourceConfig:
        if model:
            return self.registry[model._meta.label]
        if model_label:
            return self.registry[model_label]

        raise AttributeError(_("Must include either model or model_label arg."))


def get_registry() -> ResourceRegistry:
    from django.conf import settings

    if hasattr(settings, "ANSIBLE_BASE_RESOURCE_CONFIG_MODULE"):
        from django.utils.module_loading import import_string

        resource_list = import_string(settings.ANSIBLE_BASE_RESOURCE_CONFIG_MODULE + ".RESOURCE_LIST")
        api_config = import_string(settings.ANSIBLE_BASE_RESOURCE_CONFIG_MODULE + ".APIConfig")

        return ResourceRegistry(resource_list, api_config())
    else:
        return False



import json
import re
from functools import reduce

from django.contrib.contenttypes.fields import GenericForeignKey
from django.core.exceptions import FieldDoesNotExist, FieldError, ValidationError
from django.db import models
from django.db.models import BooleanField, CharField, IntegerField, JSONField, Q, TextField
from django.db.models.fields.related import ForeignKey, ForeignObjectRel, ManyToManyField
from django.db.models.functions import Cast
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import ParseError
from rest_framework.filters import BaseFilterBackend

from ansible_base.lib.utils.validation import to_python_boolean
from ansible_base.rest_filters.utils import get_fields_from_path


class FieldLookupBackend(BaseFilterBackend):
    """
    Filter using field lookups provided via query string parameters.
    """

    SUPPORTED_LOOKUPS = (
        'exact',
        'iexact',
        'contains',
        'icontains',
        'startswith',
        'istartswith',
        'endswith',
        'iendswith',
        'regex',
        'iregex',
        'gt',
        'gte',
        'lt',
        'lte',
        'in',
        'isnull',
        'search',
    )

    # A list of fields that we know can be filtered on without the possibility
    # of introducing duplicates
    NO_DUPLICATES_ALLOW_LIST = (CharField, IntegerField, BooleanField, TextField)

    # If True, JSONField will be treated as a text field for filtering purposes
    # True by default to maintain backwards compatibility
    TREAT_JSONFIELD_AS_TEXT = True

    def get_fields_from_lookup(self, model, lookup):
        if '__' in lookup and lookup.rsplit('__', 1)[-1] in self.SUPPORTED_LOOKUPS:
            path, suffix = lookup.rsplit('__', 1)
        else:
            path = lookup
            suffix = 'exact'

        if not path:
            raise ParseError(_('Query string field name not provided.'))

        # FIXME: Could build up a list of models used across relationships, use
        # those lookups combined with request.user.get_queryset(Model) to make
        # sure user cannot query using objects he could not view.
        field_list, new_path = get_fields_from_path(model, path, treat_jsonfield_as_text=self.TREAT_JSONFIELD_AS_TEXT)

        new_lookup = new_path
        new_lookup = '__'.join([new_path, suffix])
        return field_list, new_lookup

    def get_field_from_lookup(self, model, lookup):
        '''Method to match return type of single field, if needed.'''
        field_list, new_lookup = self.get_fields_from_lookup(model, lookup)
        return (field_list[-1], new_lookup)

    def to_python_related(self, value):
        value = force_str(value)
        if value.lower() in ('none', 'null'):
            return None
        else:
            return int(value)

    def value_to_python_for_field(self, field, value):
        if isinstance(field, models.BooleanField):
            return to_python_boolean(value)
        elif isinstance(field, (ForeignObjectRel, ManyToManyField, GenericForeignKey, ForeignKey)):
            try:
                return self.to_python_related(value)
            except ValueError:
                field_name = getattr(field, 'name', 'related field')
                raise ParseError(_("Invalid %(field_name)s id: %(field_id)s") % {"field_name": field_name, "field_id": value})
        else:
            return field.to_python(value)

    def value_to_python(self, model, lookup, value):
        try:
            lookup.encode("ascii")
        except UnicodeEncodeError:
            raise ValueError(_("%(lookup)s is not an allowed field name. Must be ascii encodable.") % {"lookup": lookup})

        field_list, new_lookup = self.get_fields_from_lookup(model, lookup)
        field = field_list[-1]

        needs_distinct = not all(isinstance(f, self.NO_DUPLICATES_ALLOW_LIST) for f in field_list)

        # Type names are stored without underscores internally, but are presented and
        # and serialized over the API containing underscores so we remove `_`
        # for polymorphic_ctype__model lookups.
        if new_lookup.startswith('polymorphic_ctype__model'):
            value = value.replace('_', '')
        elif new_lookup.endswith('__isnull'):
            value = to_python_boolean(value)
        elif new_lookup.endswith('__in'):
            items = []
            if not value:
                raise ValueError(_('cannot provide empty value for __in'))
            for item in value.split(','):
                items.append(self.value_to_python_for_field(field, item))
            value = items
        elif new_lookup.endswith('__regex') or new_lookup.endswith('__iregex'):
            try:
                re.compile(value)
            except re.error as e:
                raise ValueError(e.args[0])
        elif new_lookup.endswith('__iexact'):
            if not isinstance(field, (CharField, TextField)) and not (isinstance(field, JSONField) and not self.TREAT_JSONFIELD_AS_TEXT):
                raise ValueError(_('%(field_name)s is not a text field and cannot be filtered by case-insensitive search') % {'field_name': field.name})
        elif new_lookup.endswith('__search'):
            related_model = getattr(field, 'related_model', None)
            if not related_model:
                raise ValueError(_('%(related_model)s is not searchable') % {"related_model": new_lookup[:-8]})
            new_lookups = []
            for rm_field in related_model._meta.fields:
                if rm_field.name in ('username', 'first_name', 'last_name', 'email', 'name', 'description', 'playbook'):
                    new_lookups.append('{}__{}__icontains'.format(new_lookup[:-8], rm_field.name))
            return value, new_lookups, needs_distinct
        else:
            if self.TREAT_JSONFIELD_AS_TEXT and isinstance(field, JSONField):
                new_lookup = new_lookup.replace(field.name, f'{field.name}_as_txt')
            value = self.value_to_python_for_field(field, value)
        return value, new_lookup, needs_distinct

    def reserved_names(self, view):
        """The names in query_params to ignore given the current settings and current view"""
        from django.conf import settings

        reserved_set = set(settings.ANSIBLE_BASE_REST_FILTERS_RESERVED_NAMES)

        if hasattr(view, 'rest_filters_reserved_names'):
            reserved_set |= set(view.rest_filters_reserved_names)

        return reserved_set

    def filter_queryset(self, request, queryset, view):
        try:
            # Apply filters specified via query_params. Each entry in the lists
            # below is (negate, field, value).
            and_filters = []
            or_filters = []
            chain_filters = []
            role_filters = []
            search_filters = {}
            needs_distinct = False
            # Can only have two values: 'AND', 'OR'
            # If 'AND' is used, an item must satisfy all conditions to show up in the results.
            # If 'OR' is used, an item just needs to satisfy one condition to appear in results.
            search_filter_relation = 'OR'
            for key, values in request.query_params.lists():
                if key in self.reserved_names(view):
                    continue

                # HACK: make `created` available via API for the Django User ORM model
                # so it keep compatibility with other objects which exposes the `created` attr.
                if queryset.model._meta.object_name == 'User' and key.startswith('created'):
                    key = key.replace('created', 'date_joined')

                # TODO: Do we want to keep these AWX specific items here?
                # HACK: Make job event filtering by host name mostly work even
                # when not capturing job event hosts M2M.
                if queryset.model._meta.object_name == 'JobEvent' and key.startswith('hosts__name'):
                    key = key.replace('hosts__name', 'or__host__name')
                    or_filters.append((False, 'host__name__isnull', True))

                # Custom __int filter suffix (internal use only).
                q_int = False
                if key.endswith('__int'):
                    key = key[:-5]
                    q_int = True

                # RBAC filtering
                if key == 'role_level':
                    role_filters.append(values[0])
                    continue

                # Search across related objects.
                if key.endswith('__search'):
                    if values and ',' in values[0]:
                        search_filter_relation = 'AND'
                        values = reduce(lambda list1, list2: list1 + list2, [i.split(',') for i in values])
                    for value in values:
                        search_value, new_keys, _ = self.value_to_python(queryset.model, key, force_str(value))
                        assert isinstance(new_keys, list)
                        search_filters[search_value] = new_keys
                    # by definition, search *only* joins across relations,
                    # so it _always_ needs a .distinct()
                    needs_distinct = True
                    continue

                # Custom chain__ and or__ filters, mutually exclusive (both can
                # precede not__).
                q_chain = False
                q_or = False
                if key.startswith('chain__'):
                    key = key[7:]
                    q_chain = True
                elif key.startswith('or__'):
                    key = key[4:]
                    q_or = True

                # Custom not__ filter prefix.
                q_not = False
                if key.startswith('not__'):
                    key = key[5:]
                    q_not = True

                # Convert value(s) to python and add to the appropriate list.
                for value in values:
                    if q_int:
                        value = int(value)
                    value, new_key, distinct = self.value_to_python(queryset.model, key, value)
                    if distinct:
                        needs_distinct = True
                    if '_as_txt' in new_key:
                        fname = next(item for item in new_key.split('__') if item.endswith('_as_txt'))
                        queryset = queryset.annotate(**{fname: Cast(fname[:-7], output_field=TextField())})
                    if q_chain:
                        chain_filters.append((q_not, new_key, value))
                    elif q_or:
                        or_filters.append((q_not, new_key, value))
                    else:
                        and_filters.append((q_not, new_key, value))

            # Now build Q objects for database query filter.
            if and_filters or or_filters or chain_filters or role_filters or search_filters:
                args = []
                for n, k, v in and_filters:
                    if n:
                        args.append(~Q(**{k: v}))
                    else:
                        args.append(Q(**{k: v}))
                for role_name in role_filters:
                    if not hasattr(queryset.model, 'accessible_pk_qs'):
                        raise ParseError(_('Cannot apply role_level filter to this list because its model does not use roles for access control.'))
                    args.append(Q(pk__in=queryset.model.accessible_pk_qs(request.user, role_name)))
                if or_filters:
                    q = Q()
                    for n, k, v in or_filters:
                        if n:
                            q |= ~Q(**{k: v})
                        else:
                            q |= Q(**{k: v})
                    args.append(q)
                if search_filters and search_filter_relation == 'OR':
                    q = Q()
                    for term, constrains in search_filters.items():
                        for constrain in constrains:
                            q |= Q(**{constrain: term})
                    args.append(q)
                elif search_filters and search_filter_relation == 'AND':
                    for term, constrains in search_filters.items():
                        q_chain = Q()
                        for constrain in constrains:
                            q_chain |= Q(**{constrain: term})
                        queryset = queryset.filter(q_chain)
                for n, k, v in chain_filters:
                    if n:
                        q = ~Q(**{k: v})
                    else:
                        q = Q(**{k: v})
                    queryset = queryset.filter(q)
                queryset = queryset.filter(*args)
                if needs_distinct:
                    queryset = queryset.distinct()
            return queryset
        except (FieldError, FieldDoesNotExist, ValueError, TypeError) as e:
            raise ParseError(e.args[0]) from e
        except ValidationError as e:
            raise ParseError(json.dumps(e.messages, ensure_ascii=False))



from django.contrib.auth.models import AbstractUser, UserManager


class AbstractDABUser(AbstractUser):
    class Meta(AbstractUser.Meta):
        abstract = True

    all_objects = UserManager()
    objects = UserManager()



from django.conf import settings
from rest_framework.permissions import SAFE_METHODS, BasePermission


class OAuth2TokenPermission(BasePermission):
    """
    An app token is a token that has an application attached to it
    A personal access token (PAT) is a token with no application attached to it
    With that in mind:
    - An app token can be read, changed, or deleted if:
      - I am the superuser
      - I am the admin of the organization of the application of the token
      - I am the user of the token
    - An app token can be created if:
      - I have read access to the application (currently this means: I am the superuser)
    - A PAT can be read, changed, or deleted if:
      - I am the superuser
      - I am the user of the token
    - A PAT can be created if:
      - I am a user
    """

    def has_permission(self, request, view):
        # Handle PAT and app token creation separately
        if request.method == 'POST':
            if request.data.get('application'):
                # TODO: Change this once ansible/django-ansible-base#424 is fixed
                return request.user.is_superuser
        return request.user.is_authenticated

    def has_object_permission(self, request, view, obj):
        if request.method in SAFE_METHODS and getattr(request.user, 'is_platform_auditor', False):
            return True
        if request.user.is_superuser:
            return True
        if 'ansible_base.rbac' in settings.INSTALLED_APPS:
            if obj.application and obj.application.organization.access_qs(request.user, "change").exists():
                return True
        return request.user == obj.user



# Generated by Django 4.2.17 on 2025-02-25 14:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dab_oauth2_provider', '0007_alter_oauth2accesstoken_application_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='oauth2application',
            name='app_url',
            field=models.URLField(blank=True, default=None, help_text='The URL of this application.', null=True),
        ),
    ]



import logging

from django.db import IntegrityError, transaction
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers
from rest_framework.fields import empty
from rest_framework.serializers import ValidationError

from ansible_base.authentication.models import Authenticator
from ansible_base.lib.serializers.fields import JSONField
from ansible_base.lib.utils.response import get_relative_url

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.base')


def _field_required(field):
    if hasattr(field, 'required'):
        return field.required
    elif hasattr(field, 'allow_null'):
        return not field.allow_null
    return True


class BaseAuthenticatorConfiguration(serializers.Serializer):
    documentation_url = None
    ADDITIONAL_UNVERIFIED_ARGS = JSONField(
        help_text=_("Any additional fields that this authenticator can take. They are not validated and passed directly back to the authenticator."),
        required=False,
        allow_null=True,
        ui_field_label=_('Additional Authenticator Fields'),
    )

    def get_configuration_schema(self):
        fields = self.get_fields()

        schema = []

        for f in fields:
            field = fields[f]
            default = None
            if field.default is not empty:
                default = field.default

            schema_data = {
                "name": f,
                "help_text": field.help_text,
                "required": _field_required(field),
                "default": default,
                "type": field.__class__.__name__,
                "ui_field_label": getattr(field, 'ui_field_label', _('Undefined')),
            }
            if getattr(field, 'choices', None):
                schema_data["choices"] = getattr(field, 'choices')

            schema.append(schema_data)
        return schema


class AbstractAuthenticatorPlugin:
    """
    Base class for non social auth backends
    """

    configuration_class = BaseAuthenticatorConfiguration
    configuration_encrypted_fields = []

    def __init__(self, database_instance=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.database_instance = database_instance

    def set_logger(self, logger) -> None:
        if not logger:
            self.logger = logging.getLogger('ansible_base.authentication.models.abstract_authenticator')
        else:
            self.logger = logger

    def validate_configuration(self, data: dict, instance: object) -> dict:
        if not issubclass(self.configuration_class, BaseAuthenticatorConfiguration):
            raise TypeError(_("self.configuration_class must subclass BaseAuthenticatorConfiguration."))

        serializer = self.configuration_class(data=data, instance=instance)
        serializer.is_valid(raise_exception=True)

        allowed_fields = serializer.get_fields()
        errors = {}
        for k in data:
            if k not in allowed_fields:
                errors[k] = _(f"{k} is not a supported configuration option.")

        if errors:
            raise ValidationError(errors)

        return serializer.validated_data

    def to_representation(self, instance: object):
        if not issubclass(self.configuration_class, BaseAuthenticatorConfiguration):
            raise TypeError("self.configuration_class must subclass BaseAuthenticatorConfiguration.")
        serializer = self.configuration_class(data=instance.configuration, instance=instance)
        response = serializer.to_representation(instance.configuration)
        return response

    def update_settings(self, database_authenticator: Authenticator) -> None:
        self.settings = database_authenticator.configuration

    def update_if_needed(self, database_authenticator: Authenticator) -> None:
        if not self.database_instance or self.database_instance.modified != database_authenticator.modified:
            if self.database_instance:
                self.logger.info(f"Updating {self.type} adapter {database_authenticator.name}")
            else:
                self.logger.info(f"Creating an {self.type} adapter from {database_authenticator.name}")
            self.database_instance = database_authenticator
            self.update_settings(database_authenticator)
        else:
            self.logger.info(f"No updated needed for {self.type} adapter {database_authenticator.name}")

    def get_default_attributes(self):
        """
        Each backend must return a list of common attributes that are available for the authenticator map.
        These values will  be queryable by the API so that the UI can help the user configure authenticator maps.

        This list won't be comprehensive since we may not know what's available until a user logs in.

        Users will be able to configure attributes which they know exist in the Authenticator model. Additionally,
        the list of available attributes returned to the api should include fields in AuthenticatorUser.extra
        once user's have started logging in with the authenticator.
        """
        raise NotImplementedError("Implement in subclass.")

    def get_login_url(self, authenticator):
        if authenticator.category == 'sso':
            return get_relative_url('social:begin', kwargs={'backend': authenticator.slug})

    def add_related_fields(self, request, authenticator):
        return {}

    def validate(self, serializer, data):
        return data

    def move_authenticator_user_to(self, new_user, old_authenticator_user):
        """
        new_user: django User instance. User that we're moving this account to.
        old_authenticator_user: AuthenticatorUser instance from this authenticator that is being removed.
        """
        exclude_fields = (
            "social_auth",
            "authenticator_users",
            "groups",
            "has_roles",
            # We're ignoring role assignments for two reasons: 1. this isn't safe to copy right now, as it
            # could break the caching layer, 2. roles are intented to come from the authenticator via an
            # authenticator map, so when a user is move to a new authenticator, they're old roles should
            # be removed.
            "role_assignments",
            "logentry",
        )

        old_user = old_authenticator_user.user

        # Delete the old authenticator user
        old_authenticator_user.delete()

        if new_user.pk == old_user.pk:
            return

        # Copy all of the relationships from the old user to the new one
        for field in new_user._meta.get_fields():
            if field.many_to_many is True or field.one_to_many is True:
                name = field.name
                if name in exclude_fields:
                    continue
                if not hasattr(old_user, name) or not hasattr(new_user, name):
                    continue
                for x in getattr(old_user, name).all():
                    # The only case where this might fail is if the relationship has a uniqueness
                    # contraint on the user. In this case, all we can do is skip.
                    try:
                        # This atomic block is here to prevent a failure that is best described by
                        # this stack overflow: https://stackoverflow.com/questions/21458387
                        with transaction.atomic():
                            getattr(new_user, name).add(x)
                    except IntegrityError as e:
                        logger.warning(f"Could not add {name} to {new_user.username}. Error: {e}")
                        continue

        return old_user

    def get_alternative_uid(self, **kwargs):
        """
        This method can be used to provide an alternative UID for the user in case we need to match
        UIDs across different authenticators (as is the case for auto_migrate_users_to). It receives
        the kwargs from the social auth pipeline (https://python-social-auth.readthedocs.io/en/latest/pipeline.html).

        For a good example of this method in action, check out the keycloak authenticator.
        """
        return None



from .authenticator import AuthenticatorViewSet  # noqa: F401
from .authenticator_map import AuthenticatorMapViewSet  # noqa: F401
from .authenticator_plugins import AuthenticatorPluginView  # noqa: F401
from .trigger_definition import TriggerDefinitionView  # noqa: F401
from .ui_auth import UIAuth  # noqa: F401



import logging

from django.conf import settings
from django.utils.translation import gettext_lazy as _
from rest_framework.settings import api_settings, import_from_string

from ansible_base.lib.utils.settings import get_from_import
from ansible_base.lib.utils.views.ansible_base import AnsibleBaseView

logger = logging.getLogger('ansible_base.lib.utils.views.django_app_api')


# Determine and load the parent view
# If anything fails in here its a pretty low level exception so we don't catch anything and just let it raise
parent_view = getattr(settings, 'ANSIBLE_BASE_CUSTOM_VIEW_PARENT', None)
parent_view_class = AnsibleBaseView
if parent_view:
    try:
        module_name, junk, class_name = parent_view.rpartition('.')
        logger.debug(_("Trying to import parent view {class_name} from package {module_name}".format(class_name=class_name, module_name=module_name)))
        if not module_name or not class_name:
            logger.error(_("ANSIBLE_BASE_CUSTOM_VIEW_PARENT must be in the format package.subpackage.view, defaulting to AnsibleBaseView"))
        else:
            parent_view_class = get_from_import(module_name, class_name)
    except ModuleNotFoundError:
        logger.error(_("Failed to find parent view class {parent_view}, defaulting to AnsibleBaseView".format(parent_view=parent_view)))
    except ImportError:
        logger.exception(_("Failed to import {parent_view}, defaulting to AnsibleBaseView".format(parent_view=parent_view)))


# In case an app is using features like authentication, resources, or RBAC
# but not using rest_filters app, we need to specify the same filter backends
dab_filter_backends = list(api_settings.DEFAULT_FILTER_BACKENDS)

for backend_name in getattr(settings, 'ANSIBLE_BASE_CUSTOM_VIEW_FILTERS', ()):
    backend = import_from_string(backend_name, 'ANSIBLE_BASE_CUSTOM_VIEW_FILTERS')
    if backend not in dab_filter_backends:
        dab_filter_backends.append(backend)


class AnsibleBaseDjangoAppApiView(parent_view_class):
    filter_backends = dab_filter_backends



# Generated by Django 4.2.11 on 2024-05-08 14:27

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import oauth2_provider.generators
import oauth2_provider.models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dab_oauth2_provider', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='oauth2refreshtoken',
            options={'ordering': ('id',), 'verbose_name': 'refresh token'},
        ),
        migrations.RemoveField(
            model_name='oauth2accesstoken',
            name='updated',
        ),
        migrations.RemoveField(
            model_name='oauth2application',
            name='updated',
        ),
        migrations.RemoveField(
            model_name='oauth2idtoken',
            name='updated',
        ),
        migrations.RemoveField(
            model_name='oauth2refreshtoken',
            name='updated',
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='application',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='access_tokens', to=settings.OAUTH2_PROVIDER_APPLICATION_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created'),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='expires',
            field=models.DateTimeField(),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='id',
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created'),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='modified_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='scope',
            field=models.CharField(blank=True, choices=[('read', 'Read'), ('write', 'Write')], default='write', help_text="Allowed scopes, further restricts user's permissions. Must be a simple space-separated string with allowed scopes ['read', 'write'].", max_length=32),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='token',
            field=models.CharField(max_length=255, unique=True),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='user',
            field=models.ForeignKey(blank=True, help_text='The user representing the token owner', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='access_tokens', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='client_secret',
            field=oauth2_provider.models.ClientSecretField(blank=True, db_index=True, default=oauth2_provider.generators.generate_client_secret, help_text='Hashed on Save. Copy it now if this is a new secret.', max_length=255),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created'),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='id',
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created'),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='modified_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='name',
            field=models.CharField(help_text='The name of this resource', max_length=512),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='applications', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2idtoken',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created'),
        ),
        migrations.AlterField(
            model_name='oauth2idtoken',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2idtoken',
            name='expires',
            field=models.DateTimeField(),
        ),
        migrations.AlterField(
            model_name='oauth2idtoken',
            name='id',
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='oauth2idtoken',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created'),
        ),
        migrations.AlterField(
            model_name='oauth2idtoken',
            name='modified_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='application',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.OAUTH2_PROVIDER_APPLICATION_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='created',
            field=models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created'),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='created_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who created this resource', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_created+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='id',
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The date/time this resource was created'),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='modified_by',
            field=models.ForeignKey(default=None, editable=False, help_text='The user who last modified this resource', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='%(app_label)s_%(class)s_modified+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='token',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='%(app_label)s_%(class)s', to=settings.AUTH_USER_MODEL),
        ),
    ]



import logging
from typing import Optional

from crum import get_current_request
from django.http import HttpRequest

from ansible_base.jwt_consumer.common.util import validate_x_trusted_proxy_header
from ansible_base.lib.utils.settings import get_setting

logger = logging.getLogger('ansible_base.lib.uitls.requests')


def get_remote_host(request: HttpRequest) -> Optional[str]:
    value = get_remote_hosts(request, get_first_only=True)
    return value[0] if value else None


def split_header(value: str) -> list[str]:
    values = []
    for a_value in value.split(','):
        a_value = a_value.strip()
        if a_value:
            values.append(a_value)
    return values


def get_remote_hosts(request: HttpRequest, get_first_only: bool = False) -> list[str]:
    '''
    Get all IPs from the allowed headers
    NOTE: this function does not unique the hosts to preserve the order of hosts found in the variables
    '''
    remote_hosts = []

    if not request or not hasattr(request, 'META'):
        return remote_hosts

    headers = get_setting('REMOTE_HOST_HEADERS', ['REMOTE_ADDR', 'REMOTE_HOST'])

    for header in headers:
        for value in split_header(request.META.get(header, '')):
            remote_hosts.append(value)

    # If we are connected to from a trusted proxy then we can add some additional headers
    try:
        if 'HTTP_X_TRUSTED_PROXY' in request.META:
            if validate_x_trusted_proxy_header(request.META['HTTP_X_TRUSTED_PROXY']):
                # The last entry in x-forwarded-for from envoy can be trusted implicitly
                # https://www.envoyproxy.io/docs/envoy/latest/configuration/http/http_conn_man/headers#x-forwarded-for
                values = split_header(request.META.get('HTTP_X_FORWARDED_FOR', ''))
                if values:
                    remote_hosts.insert(0, values[-1])

                # x-envoy-external-address can always be trusted when coming from envoy
                # https://www.envoyproxy.io/docs/envoy/latest/configuration/http/http_conn_man/headers#x-envoy-external-address
                for value in reversed(split_header(request.META.get('HTTP_X_ENVOY_EXTERNAL_ADDRESS', ''))):
                    remote_hosts.insert(0, value)
            else:
                logger.error("Unable to use headers from trusted proxy because shared secret was invalid!")
    except Exception:
        logger.exception("Failed to validate HTTP_X_TRUSTED_PROXY")

    if get_first_only and len(remote_hosts) > 0:
        return [remote_hosts[0]]

    return remote_hosts


def is_proxied_request(request: Optional[HttpRequest] = None) -> bool:
    "Return true if request claims to be from a proxy and the header validates as such."
    if request is None:
        request = get_current_request()
    if request is None:
        # e.g. being called by CLI or something
        return False
    if x_trusted_proxy := request.META.get("HTTP_X_TRUSTED_PROXY"):
        return validate_x_trusted_proxy_header(x_trusted_proxy)
    return False



from django.conf import settings
from django.shortcuts import redirect

from ansible_base.resource_registry.resource_server import get_resource_server_config
from ansible_base.resource_registry.utils.auth_code import get_user_auth_code
from ansible_base.resource_registry.utils.settings import resource_server_defined


def redirect_to_resource_server(*args, social=None, user=None, **kwargs):
    """
    This MUST come at the end of the SOCIAL_AUTH_PIPELINE configuration.
    """

    # Allow for disabling this pipeline without removing it from the settings.
    # If resource server is defined, also silently quit
    # for ease of connected vs disconnected configs
    if (not getattr(settings, 'ENABLE_SERVICE_BACKED_SSO', False)) or (not resource_server_defined()):
        return None

    oidc_alt_key = None

    # Galaxy and AWX use different social auth backends for keycloak. AWX uses the
    # generic "oidc" provider, whereas Galaxy uses the "keycloak" provider. The way
    # these two backends handle the social auth UID is slightly different. The generic
    # backend uses the "sub" keyword in the ID token keycloak one uses the "preferred_username".
    # To be able to automatically link up accounts from these two services, we have
    # a field called "oidc_alt_key" in our auth code which is used to provide an
    # alternative lookup mechanism for the SSO user. If "sub" is used for the UID,
    # we'll pass "preferred_username" to oidc_alt_key, otherwise this gets set to "sub".
    if response := kwargs.get("response"):
        sub = response.get("sub", None)
        username = response.get("preferred_username", None)

        if sub == social.uid:
            oidc_alt_key = username
        else:
            oidc_alt_key = sub

    if not user:
        return None

    redirect_path = getattr(
        settings,
        'SERVICE_BACKED_SSO_AUTH_CODE_REDIRECT_PATH',
        "/api/gateway/v1/legacy_auth/authenticate_sso/",
    ).strip("/")

    redirect_url = getattr(
        settings,
        'SERVICE_BACKED_SSO_AUTH_CODE_REDIRECT_URL',
        get_resource_server_config()["URL"],
    )

    auth_code = get_user_auth_code(user, social_user=social, oidc_alt_key=oidc_alt_key)
    url = f"{redirect_url}/{redirect_path}/?auth_code={auth_code}"

    return redirect(url, permanent=False)



import logging
from collections import OrderedDict

from django.conf import settings
from django.db.models import Q
from django.http import HttpResponseNotFound
from django.shortcuts import get_object_or_404
from rest_framework import permissions
from rest_framework.decorators import action
from rest_framework.pagination import PageNumberPagination
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet, mixins

from ansible_base.lib.utils.response import CSVStreamResponse, get_relative_url
from ansible_base.lib.utils.views.django_app_api import AnsibleBaseDjangoAppApiView
from ansible_base.lib.utils.views.permissions import try_add_oauth2_scope_permission
from ansible_base.resource_registry.models import Resource, ResourceType, service_id
from ansible_base.resource_registry.registry import get_registry
from ansible_base.resource_registry.serializers import ResourceListSerializer, ResourceSerializer, ResourceTypeSerializer, UserAuthenticationSerializer
from ansible_base.resource_registry.utils.auth_code import get_user_auth_code
from ansible_base.rest_filters.rest_framework.field_lookup_backend import FieldLookupBackend
from ansible_base.rest_filters.rest_framework.order_backend import OrderByBackend
from ansible_base.rest_filters.rest_framework.type_filter_backend import TypeFilterBackend

logger = logging.getLogger('ansible_base.resource_registry.views')


class HasResourceRegistryPermissions(permissions.BasePermission):
    def has_permission(self, request, view):
        user = request.user

        if user.is_superuser:
            return True

        if allowed_actions := getattr(user, "resource_api_actions", None):
            if allowed_actions == "*":
                return True
            else:
                if hasattr(view, 'action'):
                    return view.action in allowed_actions
                elif hasattr(view, 'custom_action_label'):
                    return view.custom_action_label in allowed_actions
                else:
                    logger.warning(f'View {view} denied request because view action can not be identified')

        return False


class ResourcesPagination(PageNumberPagination):
    # PageNumberPagination by itself doesn't work in some apps because when api_settings.PAGE_SIZE
    # isn't set, the default is no pagination.
    page_size = 50


class ResourceAPIMixin:
    """
    The resource API is not intended to be consistent with the REST API on the service
    that it is hosted on. It is only intended to be consistent with itself. The point
    of the resource API is to provide the exact same interface on every single AAP service.
    To that end, we are not using any of the default DRF configurations for these views,
    rather we overriding all of them in order to provide the same experience everywhere.
    Regardless of where the ResourceAPI is served from it must:

    - Use DAB filters
    - Validate user access based on the AAP JWT token
    - Use Page/Number pagination
    """

    filter_backends = (FieldLookupBackend, TypeFilterBackend, OrderByBackend)
    permission_classes = try_add_oauth2_scope_permission(
        [
            HasResourceRegistryPermissions,
        ]
    )
    pagination_class = ResourcesPagination


class ResourceViewSet(
    ResourceAPIMixin,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
    GenericViewSet,
    AnsibleBaseDjangoAppApiView,
):
    """
    Index of all the resources in the system.
    """

    queryset = Resource.objects.select_related("content_type__resource_type").all()
    serializer_class = ResourceSerializer
    lookup_field = "ansible_id"

    def get_serializer_class(self):
        if self.action == "list":
            return ResourceListSerializer

        return super().get_serializer_class()

    def perform_destroy(self, instance):
        instance.delete_resource()


class ResourceTypeViewSet(
    ResourceAPIMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    GenericViewSet,
    AnsibleBaseDjangoAppApiView,
):
    """
    Index of the resource types that are configured in the system.
    """

    queryset = ResourceType.objects.all()
    serializer_class = ResourceTypeSerializer
    lookup_field = "name"
    lookup_value_regex = "[^/]+"

    def serialize_resources_hashes(self, resources_qs, serializer_class):
        """A generator that yields str sequences for csv stream response"""
        yield ("ansible_id", "resource_hash")
        for resource in resources_qs:
            yield (resource.ansible_id, serializer_class(resource.content_object).get_hash())

    @action(detail=True, methods=["get"])
    def manifest(self, request, name, *args, **kwargs):
        """
        Returns the as a stream the csv of resource_id,hash for a given resource type.
        """
        resource_type = get_object_or_404(ResourceType, name=name)
        if not resource_type.serializer_class:  # pragma: no cover
            return HttpResponseNotFound()

        if 'service_id' in request.query_params:
            if request.query_params['service_id'] == 'all':
                service_filter = Q()
            else:
                # Return this services resources plus the resources from the service requested
                service_filter = Q(service_id=service_id()) | Q(service_id=request.query_params['service_id'])
        else:
            service_filter = Q(service_id=service_id())

        resources = Resource.objects.filter(content_type__resource_type=resource_type).filter(service_filter).prefetch_related("content_object")

        if name == "shared.user" and (system_user := getattr(settings, "SYSTEM_USERNAME", None)):
            resources = resources.exclude(name=system_user)

        if not resources:
            return HttpResponseNotFound()

        return CSVStreamResponse(self.serialize_resources_hashes(resources, resource_type.serializer_class)).stream()


class ServiceMetadataView(
    AnsibleBaseDjangoAppApiView,
):
    permission_classes = try_add_oauth2_scope_permission(
        [
            HasResourceRegistryPermissions,
        ]
    )

    # Corresponds to viewset action but given a different name so schema generators are not messed up
    custom_action_label = "service-metadata"

    def get(self, request, **kwargs):
        registry = get_registry()
        return Response({"service_id": service_id(), "service_type": registry.api_config.service_type})


class ServiceIndexRootView(AnsibleBaseDjangoAppApiView):
    permission_classes = try_add_oauth2_scope_permission([permissions.IsAuthenticated])

    def get(self, request, format=None):
        '''Link other resource registry endpoints'''

        data = OrderedDict()
        data['metadata'] = get_relative_url('service-metadata')
        data['resources'] = get_relative_url('resource-list')
        data['resource-types'] = get_relative_url('resourcetype-list')
        return Response(data)


class ValidateLocalUserView(AnsibleBaseDjangoAppApiView):
    """
    Validate a user's username and password.
    """

    custom_action_label = "validate-local-user"

    permission_classes = [AllowAny]

    def post(self, request, **kwargs):
        serializer = UserAuthenticationSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        api_config = get_registry().api_config
        user = api_config.authenticate_local_user(serializer.validated_data["username"], serializer.validated_data["password"])

        if not user:
            return Response(status=401)

        try:
            auth_code = get_user_auth_code(user)
        except AttributeError:
            logger.exception(f"Cannot generate auth code for user {user}")
            auth_code = None

        response = {
            "ansible_id": Resource.get_resource_for_object(user).ansible_id,
            "auth_code": auth_code,
        }

        return Response(data=response)



from __future__ import annotations  # support python<3.10

import asyncio
import csv
import logging
import time
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from io import StringIO, TextIOBase

from asgiref.sync import sync_to_async
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import transaction
from django.db.models import QuerySet
from django.db.utils import Error, IntegrityError
from requests import HTTPError

from ansible_base.resource_registry.models import Resource, ResourceType
from ansible_base.resource_registry.models.service_identifier import service_id
from ansible_base.resource_registry.registry import get_registry
from ansible_base.resource_registry.rest_client import ResourceAPIClient, get_resource_server_client

logger = logging.getLogger('ansible_base.resources_api.tasks.sync')


class ManifestNotFound(HTTPError):
    """Raise when server returns 404 for a manifest"""


class ResourceDeletionError(Error):
    """Raise for deletion errors on Django ORM"""


class ResourceSyncHTTPError(HTTPError):
    """Custom catchall error"""


class SyncStatus(str, Enum):
    CREATED = "created"
    UPDATED = "updated"
    NOOP = "noop"
    CONFLICT = "conflict"
    UNAVAILABLE = "unavailable"
    ERROR = "error"


@dataclass
class ManifestItem:
    ansible_id: str
    resource_hash: str
    resource_data: dict | None = None

    def __hash__(self):
        return hash(self.ansible_id)


@dataclass
class SyncResult:
    status: SyncStatus
    item: ManifestItem

    def __iter__(self):
        """Allows unpacking  status, item = SyncResult(...)"""
        return iter((self.status, self.item))


def create_api_client() -> ResourceAPIClient:
    """Factory for pre-configured ResourceAPIClient."""
    params = {"raise_if_bad_request": False}

    if jwt_user_id := getattr(settings, "RESOURCE_JWT_USER_ID", None):
        params["jwt_user_id"] = jwt_user_id

    service_path = getattr(settings, "RESOURCE_SERVICE_PATH", None)
    if not service_path:
        raise ValueError("RESOURCE_SERVICE_PATH is not set.")
    params["service_path"] = service_path

    client = get_resource_server_client(**params)
    return client


def fetch_manifest(
    resource_type_name: str,
    api_client: ResourceAPIClient | None = None,
) -> list[ManifestItem]:
    """Fetch RESOURCE_SERVER manifest, parses the CSV and returns a list."""
    api_client = api_client or create_api_client()
    api_client.raise_if_bad_request = False  # Status check is needed

    resp_metadata = api_client.get_service_metadata()
    resp_metadata.raise_for_status()

    manifest_stream = api_client.get_resource_type_manifest(resource_type_name, filters={"service_id": service_id()})
    if manifest_stream.status_code == 404:
        msg = f"manifest for {resource_type_name} NOT FOUND."
        raise ManifestNotFound(msg)

    try:
        manifest_stream.raise_for_status()
    except HTTPError as exc:
        raise ResourceSyncHTTPError() from exc

    csv_reader = csv.DictReader(StringIO(manifest_stream.text))
    return [ManifestItem(**row) for row in csv_reader]


def get_orphan_resources(
    resource_type_name: str,
    manifest_list: list[ManifestItem],
) -> QuerySet:
    """QuerySet with orphaned managed resources to be deleted."""
    return (
        Resource.objects.filter(
            content_type__resource_type__name=resource_type_name,
        )
        .exclude(ansible_id__in=[item.ansible_id for item in manifest_list])
        .exclude(
            service_id=service_id(),
            is_partially_migrated=False,
        )
    )


def delete_resource(resource: Resource):
    """Wrapper to delete content_object and its related Resource.
    It is up to the caller to wrap it on a database transaction.
    """
    try:
        return resource.delete_resource()
    except Error as exc:  # pragma: no cover
        logger.error(f"Failed to delete resource {resource.ansible_id}. Received error: {exc}")
        raise ResourceDeletionError() from exc


def get_managed_resource(manifest_item: ManifestItem) -> Resource | None:
    """Return an instance containing the local managed resource to process."""
    return Resource.objects.filter(
        ansible_id=manifest_item.ansible_id,
    ).first()


def get_resource_type_names() -> list[str]:
    """Ordered list of registered resource types."""
    registry = get_registry()
    resources = registry.get_resources()
    return [f"shared.{rt.model._meta.model_name}" for _, rt in sorted(resources.items())]


def _handle_conflict(resource_data: dict, resource_type: ResourceType, api_client: ResourceAPIClient):
    # How might we end up with a conflict?
    # - get_orphan_resources ignores resources that are owned by the service and have
    #   is partially migrated set to false. We could be conflicting with one of those.
    # - You could rename A --> B and B --> C. If A is updated before B is, then we will
    #   end up needing to update B before we can update A.
    # - You could rename A --> B and B --> A. If this happens, there is really no recovery.
    #   Updating either would result in a conflict. We would have to set A or B to some
    #   temporary value. Hopefully this will never happen since this scenario would be
    #   incredibly difficult to pull off.

    conflict_resource = resource_type.get_conflicting_resource(resource_data)

    if conflict_resource is None:
        return

    resp = api_client.get_resource(conflict_resource.ansible_id)

    # If the conflicting resource doesn't exist on the server, just go ahead and delete it.
    # This will most likely happen with resources that are ignored by get_orphan_resources.
    if resp.status_code == 404:
        delete_resource(conflict_resource)

    # If the resource does exist, lets update it first. Hopefully this desn't also result
    # in a duplicate key error. If it does, we're cooked.
    elif resp.status_code == 200:
        data = resp.json()
        conflict_resource.update_resource(
            resource_data=data["resource_data"], service_id=data["service_id"], is_partially_migrated=data["is_partially_migrated"]
        )
    else:
        raise ResourceSyncHTTPError(f"Received a bad error code from the resource server: {resp.status_code}")


def _attempt_update_resource(
    manifest_item: ManifestItem,
    resource: Resource,
    resource_data: dict,
    api_client: ResourceAPIClient,
    **kwargs,
) -> SyncResult:
    """Try to update existing resource."""
    try:
        resource.update_resource(resource_data, partial=True, **kwargs)
    except IntegrityError:  # pragma: no cover
        # This typically means that there was a duplicate key error. To mitigate this
        # we will attempt to hanlde the conflicting resource and perform the operation
        # again.
        try:
            _handle_conflict(resource_data, resource.resource_type_obj, api_client)
            resource.update_resource(resource_data, partial=True, **kwargs)
        except (ResourceDeletionError, IntegrityError, Error, ValidationError) as e:
            logger.error(f"Failed to gracefully handle conflict for {resource_data}. Got error {e}.")
            return SyncResult(SyncStatus.CONFLICT, manifest_item)
    except (Error, ValidationError) as e:
        # Something happened with the database. We don't know what it is. Instead of failing the whole
        # sync, we'll raise an error and skip this for now.
        logger.error(f"Failed to update resource {resource.ansible_id}. Received error: {e}. Will try again on the next sync.")
        return SyncResult(SyncStatus.ERROR, manifest_item)

    return SyncResult(SyncStatus.UPDATED, manifest_item)


def _attempt_create_resource(
    manifest_item: ManifestItem,
    resource_data: dict,
    resource_type: ResourceType,
    resource_service_id: str,
    api_client: ResourceAPIClient,
) -> SyncResult:
    try:
        Resource.create_resource(
            resource_type=resource_type,
            resource_data=resource_data,
            ansible_id=manifest_item.ansible_id,
            service_id=resource_service_id,
        )
    except IntegrityError:
        # This typically means that there was a duplicate key error. To mitigate this
        # we will attempt to hanlde the conflicting resource and perform the operation
        # again.
        try:
            _handle_conflict(resource_data, resource_type, api_client)
            Resource.create_resource(
                resource_type=resource_type,
                resource_data=resource_data,
                ansible_id=manifest_item.ansible_id,
                service_id=resource_service_id,
            )
        except (ResourceDeletionError, IntegrityError, Error, ValidationError) as e:
            logger.error(f"Failed to gracefully handle conflict for {resource_data}. Got error {e}.")
            return SyncResult(SyncStatus.CONFLICT, manifest_item)
    except (Error, ValidationError) as e:
        # Something happened with the database. We don't know what it is. Instead of failing the whole
        # sync, we'll raise an error and skip this for now.
        logger.error(f"Failed to create {manifest_item.ansible_id}. Received error: {e}. Will try again on the next sync.")
        return SyncResult(SyncStatus.ERROR, manifest_item)

    return SyncResult(SyncStatus.CREATED, manifest_item)


def resource_sync(
    manifest_item: ManifestItem,
    api_client: ResourceAPIClient | None = None,
) -> SyncResult:
    """Uni-directional sync local resources from RESOURCE_SERVER resources."""
    api_client = api_client or create_api_client()
    local_managed_resource = get_managed_resource(manifest_item)
    resource_data = None
    resource_type_name = None
    resource_service_id = None
    unavailable = False  # for retry mechanism

    def set_resource_local_variables():
        """Inner caching function to avoid making unnecessary requests."""
        nonlocal resource_data
        nonlocal resource_type_name
        nonlocal resource_service_id
        nonlocal unavailable
        if resource_data is None or resource_type_name is None:
            resp = api_client.get_resource(manifest_item.ansible_id)
            if 400 <= resp.status_code < 500:  # pragma: no cover
                unavailable = True
                return
            resp.raise_for_status()
            resource_data = resp.json()["resource_data"]
            resource_type_name = resp.json()["resource_type"]
            resource_service_id = resp.json()["service_id"]

    if local_managed_resource:
        # Exists locally: Compare and Update
        sc = local_managed_resource.content_type.resource_type.serializer_class
        local_hash = sc(local_managed_resource.content_object).get_hash()
        if manifest_item.resource_hash == local_hash:
            return SyncResult(SyncStatus.NOOP, manifest_item)
        set_resource_local_variables()
        if unavailable:  # pragma: no cover
            return SyncResult(SyncStatus.UNAVAILABLE, manifest_item)
        # bind fetched resource_data for allowing reporting.
        manifest_item.resource_data = resource_data
        return _attempt_update_resource(
            manifest_item,
            local_managed_resource,
            resource_data,
            api_client,
            service_id=resource_service_id,
        )
    else:
        # New: Create it locally
        set_resource_local_variables()
        if unavailable:  # pragma: no cover
            return SyncResult(SyncStatus.UNAVAILABLE, manifest_item)
        manifest_item.resource_data = resource_data
        resource_type = ResourceType.objects.get(name=resource_type_name)
        return _attempt_create_resource(
            manifest_item=manifest_item,
            resource_data=resource_data,
            resource_type=resource_type,
            resource_service_id=resource_service_id,
            api_client=api_client,
        )


# https://docs.djangoproject.com/en/4.2/topics/async/#asgiref.sync.sync_to_async
async_resource_sync = sync_to_async(resource_sync)


@dataclass
class SyncExecutor:
    """Public Executor Implementing Sync and Async process."""

    api_client: ResourceAPIClient = field(default_factory=create_api_client)
    resource_type_names: list[str] | None = None
    retries: int = 0
    retrysleep: int = 30
    stdout: TextIOBase | None = None
    unavailable: set = field(default_factory=set)
    attempts: int = 0
    deleted_count: int = 0
    asyncio: bool = False
    results: dict = field(default_factory=lambda: defaultdict(list))

    def write(self, text: str = ""):
        """Write to assigned IO or simply ignores the text."""
        if self.stdout:
            self.stdout.write(text)

    def _report_manifest_item(self, result: SyncResult):
        """Record status for each single resource of the manifest."""
        msg = f"{result.status.value.upper()} {result.item.ansible_id}"
        if result.item.resource_data:
            details = result.item.resource_data.get(
                "name",
                result.item.resource_data.get("username", ""),
            )
            msg += f" {details}"
        self.write(msg)

    def _report_results(self, results: list[SyncResult]):
        """Grouped results report at the end of the execution."""
        created_count = updated_count = conflicted_count = skipped_count = error_count = 0
        for status, manifest_item in results:
            self.results[status.value].append(manifest_item)
            self.unavailable.discard(manifest_item)
            # when python>3.10 replace with match
            if status == SyncStatus.UNAVAILABLE:  # pragma: no cover
                self.unavailable.add(manifest_item)
            elif status == SyncStatus.CREATED:
                created_count += 1
            elif status == SyncStatus.UPDATED:
                updated_count += 1
            elif status == SyncStatus.CONFLICT:
                conflicted_count += 1
            elif status == SyncStatus.NOOP:
                skipped_count += 1
            elif status == SyncStatus.ERROR:
                error_count += 1
            else:  # pragma: no cover
                raise TypeError("Unhandled SyncResult")

        self.write(
            f"Processed {len(results) + self.deleted_count} | "
            f"Created {created_count} | "
            f"Updated {updated_count} | "
            f"Conflict {conflicted_count} | "
            f"Unavailable {len(self.unavailable)} | "
            f"Skipped {skipped_count} | "
            f"Deleted {self.deleted_count} | "
            f"Errors {error_count}"
        )

    async def _a_process_manifest_item(self, manifest_item):  # pragma: no cover
        """Awaitable to process a manifest item using asyncio"""
        result = await async_resource_sync(manifest_item, self.api_client)
        self._report_manifest_item(result)
        return result

    async def _a_process_manifest_list(self, manifest_list):  # pragma: no cover
        """Awaitable to process a sequence of items using Asyncio."""
        queue = [self._a_process_manifest_item(item) for item in manifest_list]
        results = await asyncio.gather(*queue)
        self._report_results(results)

    def _process_manifest_item(self, manifest_item):
        """Process a manifest item"""
        result = resource_sync(manifest_item, self.api_client)
        self._report_manifest_item(result)
        return result

    def _process_manifest_list(self, manifest_list):
        """Process items sequentially."""
        results = [self._process_manifest_item(item) for item in manifest_list]
        self._report_results(results)

    def _cleanup_orphans(self, resource_type, manifest_list):
        """Delete local managed resources that are not part of the manifest."""
        resources_to_cleanup = get_orphan_resources(
            resource_type,
            manifest_list,
        )
        self.deleted_count = resources_to_cleanup.count()
        if self.deleted_count:
            self.write(f"Deleting {self.deleted_count} orphaned resources")
            for orphan in resources_to_cleanup:
                try:
                    _sc = orphan.content_type.resource_type.serializer_class
                    data = _sc(orphan.content_object).data
                    data.update(orphan.summary_fields())
                    with transaction.atomic():
                        delete_resource(orphan)
                except ResourceDeletionError as exc:
                    self.write(f"Error deleting orphaned resources {str(exc)}")
                else:  # persist in the report
                    self.results["deleted"].append(data)

    def _handle_retries(self):  # pragma: no cover
        """Check if there are unavailable resources to re-try."""
        while self.unavailable and self.attempts < self.retries:
            self.write()
            self.write(f"Retry attempt {self.attempts}/{self.retries}")
            if self.retrysleep:
                self.write(f"waiting {self.retrysleep} seconds")
                time.sleep(self.retrysleep)
            if self.asyncio is True:
                asyncio.run(self._a_process_manifest_list(self.unavailable))
            else:
                self._process_manifest_list(self.unavailable)
            self.attempts += 1

    def _dispatch_sync_process(self, manifest_list: list[ManifestItem]):
        """Sync all the items from the manifest using either asyncio or sequentialy."""
        if self.asyncio is True:  # pragma: no cover
            self.write(f"Processing {len(manifest_list)} resources with asyncio executor.")
            self.write()
            asyncio.run(self._a_process_manifest_list(manifest_list))
        else:
            self.write(f"Processing {len(manifest_list)} resources sequentially.")
            self.write()
            self._process_manifest_list(manifest_list)

    def run(self):
        """Run the sync workflow.

        1. Iterate enabled resource types.
        2. Fetch RESOURCE_SERVER manifest.
        3. Cleanup orphaned resources (deleted remotely).
        4. Process the sync for each item in the manifest.
        5. Handle retries.
        """
        self.write("----- RESOURCE SYNC STARTED -----")
        self.write()

        for resource_type_name in get_resource_type_names():
            if self.resource_type_names and resource_type_name not in self.resource_type_names:
                # Skip types that are filtered out
                continue

            self.write(f">>> {resource_type_name}")
            try:
                manifest_list = fetch_manifest(resource_type_name, api_client=self.api_client)
            except ManifestNotFound as ex:
                self.write(str(ex))
                continue

            self._cleanup_orphans(resource_type_name, manifest_list)
            self._dispatch_sync_process(manifest_list)
            self._handle_retries()

            self.write()

        self.write("----- RESOURCE SYNC FINISHED -----")



from .access_token import OAuth2AccessToken
from .application import OAuth2Application
from .id_token import OAuth2IDToken
from .refresh_token import OAuth2RefreshToken

__all__ = (
    'OAuth2AccessToken',
    'OAuth2Application',
    'OAuth2IDToken',
    'OAuth2RefreshToken',
)

#
# There were a lot of problems making the initial migrations for this class
# See https://github.com/jazzband/django-oauth-toolkit/issues/634 which helped
#
# Here were my steps:
#  1. Make sure 'ansible_base.oauth2_provider' is in test_app INSTALLED_APPS (this already should be)
#  2. Comment out all OAUTH2_ settings in ansible_base/lib/dynamic_config/dynamic_settings.py which reference dab_oauth2_provider.*
#  3. Change all model classes to:
#         remove oauth2_models.Abstract* as superclasses (including the meta ones)
#         comment out the "import oauth2_provider.models as oauth2_models" imports
#  4. ./manage.py makemigrations dab_oauth2_provider
#  5. Edit the created 0001 migration and delete the dpendency on ('test_app', '000X')
#  6. ./manage.py migrate dab_oauth2_provider
#  7. Look at the generated migration, if this has a direct reference to your applications organization model in OAuth2Application model we need to update it
#     for example, if it looks like:
#       ('organization', ... to='<your app>.organization')),
#     We want to change this to reference the setting:
#       ('organization', ... to=settings.ANSIBLE_BASE_ORGANIZATION_MODEL)),
#     We should also add this in the migration dependencies:
#       migrations.swappable_dependency(settings.ANSIBLE_BASE_ORGANIZATION_MODEL),
#  8. Uncomment all OAUTH2_PROVIDER_* settings
#  9. Revert step 3
#  10. gateway-manage makemigrations && gateway-manage migrate ansible_base
#       When you do this django does not realize that you are creating an initial migration and tell you its impossible to migrate so fields
#       It will ask you to either: 1. Enter a default 2. Quit
#       Tell it to use the default if it has one populated at the prompt. Other wise use django.utils.timezone.now for timestamps and  '' for other items
#       This wont matter for us because there will be no data in the tables between these two migrations
#  11. You can now combine the migration into one.
#      Add the `import uuid` to the top of the initial migration file
#      Copy all of the operations from the second file to the first
#      Find the AddFields commands for oauth2refreshtoken.access_token and oauth2accesstoken.source_refresh_token and move them to the end of the operations
#      If desired, convert the remaining AddFilds into actual fields on the table creation. For example:
#           migrations.AddField(
#                model_name='oauth2accesstoken',
#                name='created',
#                field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
#                preserve_default=False,
#            ),
#      Would become the following field on the oauth2accesstoken table:
#            ('created', models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now)),
#      Next put the table creation in the following order: OAuth2Application, OAuth2IDToken, OAuth2RefreshToken, OAuth2AccessToken
#      Finally, be sure to add this to the migration file:
#            run_before = [
#              ('oauth2_provider', '0001_initial'),
#            ]
#  12. Delete the new migration
#  13. zero out the initial migration with: ./manage.py migrate dab_oauth2_provider zero
#  14. Make the actual migration with: ./manage.py migrate dab_oauth2_provider



# Generated by Django 4.2.6 on 2024-01-19 20:06

import uuid

import django.db.models.deletion
from django.db import migrations, models


def create_service_id(apps, schema_editor):
    ServiceID = apps.get_model("dab_resource_registry", "ServiceID")
    if not ServiceID.objects.exists():
        ServiceID.objects.create()


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='ServiceID',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
            ],
        ),
        migrations.RunPython(
            code=create_service_id,
            reverse_code=migrations.RunPython.noop
        ),
        migrations.CreateModel(
            name='ResourceType',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('externally_managed', models.BooleanField()),
                ('migrated', models.BooleanField(default=False)),
                ('name', models.CharField(db_index=True, editable=False, max_length=256, unique=True)),
                ('content_type', models.OneToOneField(
                    on_delete=django.db.models.deletion.CASCADE, related_name='resource_type', to='contenttypes.contenttype')),
            ],
        ),
        migrations.CreateModel(
            name='Resource',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.TextField()),
                ('service_id', models.CharField(default="00000000", max_length=8)),
                ('resource_id', models.UUIDField(db_index=True, default=uuid.uuid4, unique=True)),
                ('name', models.CharField(max_length=512, null=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='resources', to='contenttypes.contenttype')),
            ],
            options={
                'indexes': [models.Index(fields=['content_type', 'object_id'], name='dab_resourc_content_6d9d9c_idx')],
                'unique_together': {('content_type', 'object_id')},
            },
        ),
    ]



import importlib
import logging
from typing import Any

from django.conf import settings
from django.utils.translation import gettext_lazy as _

from ansible_base.lib.utils.validation import to_python_boolean

logger = logging.getLogger('ansible_base.lib.utils.settings')


class SettingNotSetException(Exception):
    pass


def get_setting(name: str, default: Any = None, log_exception: bool = True) -> Any:
    try:
        the_function = get_function_from_setting('ANSIBLE_BASE_SETTINGS_FUNCTION')
        if the_function:
            setting = the_function(name)
            return setting
    except SettingNotSetException:
        # If the setting was not set thats ok, we will fall through to trying to get it from the django setting or the default value
        pass
    except Exception:
        if log_exception:
            logger.exception(
                _(
                    'ANSIBLE_BASE_SETTINGS_FUNCTION was set but calling it as a function failed (see exception), '
                    'ignoring error and attempting to load from settings'
                )
            )

    return getattr(settings, name, default)


def get_function_from_setting(setting_name: str) -> Any:
    setting = getattr(settings, setting_name, None)
    if not setting:
        return None

    try:
        module_name, _junk, function_name = setting.rpartition('.')
        the_function = getattr(importlib.import_module(module_name), function_name)
        return the_function
    except Exception:
        logger.exception(_('{setting_name} was set but we were unable to import its reference as a function.').format(setting_name=setting_name))
        return None


def get_from_import(module_name, attr):
    "Thin wrapper around importlib.import_module, mostly exists so that we can safely mock this in tests"
    module = importlib.import_module(module_name, package=attr)
    return getattr(module, attr)


def is_aoc_instance():
    managed_cloud_setting = 'ANSIBLE_BASE_MANAGED_CLOUD_INSTALL'
    try:
        return to_python_boolean(getattr(settings, managed_cloud_setting, False))
    except ValueError:
        logger.error(f'{managed_cloud_setting} was set but could not be converted to a boolean, assuming false')
        return False



import logging
from typing import Optional, Tuple

from django.conf import settings
from django.core.cache import caches

from ansible_base.lib.utils.settings import get_setting

logger = logging.getLogger('ansible_base.jwt_consumer.common.cache')


# This setting allows a service to override which django cache we want to use
jwt_cache_name = getattr(settings, 'ANSIBLE_BASE_JWT_CACHE_NAME', 'default')
cache = caches[jwt_cache_name]
# This is the cache name we will use for the JWT key
cache_key = 'ansible_base_jwt_public_key'


class JWTCache:
    def get_cache_timeout(self):
        # If unspecified the cache will expire in 7 days
        cache_timeout = get_setting('ANSIBLE_BASE_JWT_CACHE_TIMEOUT_SECONDS', 604800)
        return cache_timeout

    def check_user_in_cache(self, validated_body: dict) -> Tuple[bool, dict]:
        # These are the defaults which will get passed to the user creation and what we expect in the cache
        expected_cache_value = {
            "first_name": validated_body['user_data']["first_name"],
            "last_name": validated_body['user_data']["last_name"],
            "email": validated_body['user_data']["email"],
            "is_superuser": validated_body['user_data']["is_superuser"],
        }
        cached_user = cache.get(validated_body["sub"], None)
        # If the user was in the cache and the values of the cache match the expected values we had it in cache
        if cached_user is not None and cached_user == expected_cache_value:
            return True, expected_cache_value
        # The user was not previously in the cache, set the user in the cache so it is found on future requests
        cache.set(validated_body["sub"], expected_cache_value, timeout=self.get_cache_timeout())
        return False, expected_cache_value

    def get_key_from_cache(self) -> Optional[str]:
        # If we are not ignoring the cache (forcing a reload of the key), check it
        key = cache.get(cache_key, None)
        logger.debug(f"Cached key is {key}")
        return key

    def set_key_in_cache(self, key: str) -> None:
        cache.set(cache_key, key, timeout=self.get_cache_timeout())



import logging

from django.urls import include, path
from rest_framework import routers

from ansible_base.resource_registry import views

logger = logging.getLogger('ansible_base.resource-urls')

service_router = routers.SimpleRouter()

service_router.register(r'resources', views.ResourceViewSet)
service_router.register(r'resource-types', views.ResourceTypeViewSet)

service = [
    path('metadata/', views.ServiceMetadataView.as_view(), name="service-metadata"),
    path('validate-local-account/', views.ValidateLocalUserView.as_view(), name="validate-local-account"),
    path('', include(service_router.urls)),
    path('', views.ServiceIndexRootView.as_view(), name='service-index-root'),
]

urlpatterns = [
    path('service-index/', include(service)),
]



import logging

from django.http import HttpResponse, HttpResponseNotFound
from django.urls import re_path
from django.utils.translation import gettext_lazy as _
from onelogin.saml2.auth import OneLogin_Saml2_Auth
from onelogin.saml2.errors import OneLogin_Saml2_Error
from onelogin.saml2.settings import OneLogin_Saml2_Settings
from rest_framework.serializers import ValidationError
from rest_framework.views import View
from social_core.backends.saml import SAMLAuth, SAMLIdentityProvider

from ansible_base.authentication.authenticator_plugins.base import AbstractAuthenticatorPlugin, BaseAuthenticatorConfiguration
from ansible_base.authentication.authenticator_plugins.utils import get_authenticator_plugin
from ansible_base.authentication.models import Authenticator
from ansible_base.authentication.social_auth import (
    AuthenticatorConfigTestStrategy,
    AuthenticatorStorage,
    AuthenticatorStrategy,
    SocialAuthMixin,
    SocialAuthValidateCallbackMixin,
)
from ansible_base.lib.serializers.fields import CharField, JSONField, ListField, PrivateKey, PublicCert, URLField
from ansible_base.lib.utils.encryption import ENCRYPTED_STRING
from ansible_base.lib.utils.response import get_relative_url
from ansible_base.lib.utils.settings import get_setting
from ansible_base.lib.utils.validation import validate_cert_with_key

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.saml')

idp_string = 'IdP'


class SAMLConfiguration(BaseAuthenticatorConfiguration):
    settings_to_enabled_idps_fields = {
        'IDP_URL': 'url',
        'IDP_X509_CERT': 'x509cert',
        'IDP_ENTITY_ID': 'entity_id',
        'IDP_ATTR_EMAIL': 'attr_email',
        'IDP_GROUPS': 'attr_groups',
        'IDP_ATTR_USERNAME': 'attr_username',
        'IDP_ATTR_LAST_NAME': 'attr_last_name',
        'IDP_ATTR_FIRST_NAME': 'attr_first_name',
        'IDP_ATTR_USER_PERMANENT_ID': 'attr_user_permanent_id',
    }

    documentation_url = "https://python-social-auth.readthedocs.io/en/latest/backends/saml.html"

    SP_ENTITY_ID = CharField(
        allow_null=False,
        max_length=512,
        default="saml_entity",
        help_text=_(
            "The application-defined unique identifier used as the audience of the SAML service provider (SP) configuration. This is usually the URL for the"
            " service."
        ),
        ui_field_label=_('SAML Service Provider Entity ID'),
    )
    SP_PUBLIC_CERT = PublicCert(
        allow_null=False,
        help_text=_("Create a keypair to use as a service provider (SP) and include the certificate content here."),
        ui_field_label=_('SAML Service Provider Public Certificate'),
    )
    SP_PRIVATE_KEY = PrivateKey(
        allow_null=False,
        help_text=_("Create a keypair to use as a service provider (SP) and include the private key content here."),
        ui_field_label=_('SAML Service Provider Private Key'),
    )
    ORG_INFO = JSONField(
        allow_null=False,
        default={"en-US": {"url": "", "name": "", "displayname": ""}},
        help_text=_("Provide the URL, display name, and the name of your app. Refer to the documentation for example syntax."),
        ui_field_label=_('SAML Service Provider Organization Info'),
    )
    TECHNICAL_CONTACT = JSONField(
        allow_null=False,
        default={'givenName': "", 'emailAddress': ""},
        help_text=_("Provide the name and email address of the technical contact for your service provider. Refer to the documentation for example syntax."),
        ui_field_label=_('SAML Service Provider Technical Contact'),
    )
    SUPPORT_CONTACT = JSONField(
        allow_null=False,
        default={'givenName': "", 'emailAddress': ""},
        help_text=_("Provide the name and email address of the support contact for your service provider. Refer to the documentation for example syntax."),
        ui_field_label=_('SAML Service Provider Support Contact'),
    )
    SP_EXTRA = JSONField(
        default={"requestedAuthnContext": False},
        help_text=_("A dict of key value pairs to be passed to the underlying python-saml Service Provider configuration setting."),
        ui_field_label=_('SAML Service Provider extra configuration data'),
    )
    SECURITY_CONFIG = JSONField(
        default={},
        help_text=_(
            "A dict of key value pairs that are passed to the underlying python-saml security setting https://github.com/onelogin/python-saml#settings."
        ),
        ui_field_label=_('SAML Security Config'),
    )
    EXTRA_DATA = ListField(
        default=[],
        help_text=_("A list of tuples that maps IDP attributes to extra_attributes. Each attribute will be a list of values, even if only 1 value."),
        ui_field_label=_('SAML IDP to extra_data attribute mapping'),
    )
    IDP_URL = URLField(
        allow_null=False,
        help_text=_("The URL to redirect the user to for login initiation."),
        ui_field_label=_('IdP Login URL'),
    )
    IDP_X509_CERT = PublicCert(
        allow_null=False,
        help_text=_("The public certificate used for secrets coming from the IdP."),
        ui_field_label=_('IdP Public Cert'),
    )
    IDP_ENTITY_ID = CharField(
        allow_null=False,
        help_text=_("The entity ID returned in the assertion."),
        ui_field_label=_('Entity ID'),
    )
    IDP_GROUPS = CharField(
        allow_null=True,
        required=False,
        help_text=_("The field in the assertion which represents the user's groups."),
        ui_field_label=_('Groups'),
    )
    IDP_ATTR_EMAIL = CharField(
        allow_null=True,
        help_text=_("The field in the assertion which represents the user's email."),
        ui_field_label=_('User Email'),
    )
    IDP_ATTR_USERNAME = CharField(
        allow_null=True,
        required=False,
        help_text=_("The field in the assertion which represents the user's username."),
        ui_field_label=_('Username'),
    )
    IDP_ATTR_LAST_NAME = CharField(
        allow_null=True,
        help_text=_("The field in the assertion which represents the user's last name."),
        ui_field_label=_('User Last Name'),
    )
    IDP_ATTR_FIRST_NAME = CharField(
        allow_null=True,
        help_text=_("The field in the assertion which represents the user's first name."),
        ui_field_label=_('User First Name'),
    )
    IDP_ATTR_USER_PERMANENT_ID = CharField(
        allow_null=True,
        required=False,
        help_text=_("The field in the assertion which represents the user's permanent id (overrides IDP_ATTR_USERNAME)"),
        ui_field_label=_('User Permanent ID'),
    )
    CALLBACK_URL = URLField(
        required=False,
        allow_null=True,
        help_text=_(
            "Register the service as a service provider (SP) with each identity provider (IdP) you have configured. "
            "Provide your SP Entity ID and this ACS URL for your application."
        ),
        ui_field_label=_('SAML Assertion Consumer Service (ACS) URL'),
    )

    def validate(self, attrs):
        # attrs is only the data in the configuration field
        errors = {}
        # pull the cert_info out of the existing object (if we have one)
        cert_info = {
            "SP_PRIVATE_KEY": getattr(self.instance, 'configuration', {}).get('SP_PRIVATE_KEY', None),
            "SP_PUBLIC_CERT": getattr(self.instance, 'configuration', {}).get('SP_PUBLIC_CERT', None),
        }

        # Get public cert if provided
        public_cert = attrs.get('SP_PUBLIC_CERT', None)
        if public_cert is not None:
            cert_info['SP_PUBLIC_CERT'] = public_cert

        # Now get the SP_PRIVATE_KEY out of the passed in attrs (if there is any)
        private_key = attrs.get('SP_PRIVATE_KEY', None)
        if private_key and private_key != ENCRYPTED_STRING:
            # We got an input form the attrs so let that override whatever was in the object
            cert_info['SP_PRIVATE_KEY'] = private_key
        # If we didn't get an input or we got ENCRYPTED_STRING but there is an item, we will just use whatever we got from the item

        # If we made it here the cert_info has one of three things:
        #  * None (error state or not passed in on PUT)
        #  * The existing value from the instance
        #  * A new value

        # Now validate that we can load the cert and key and that they match.
        # Technically, we are also doing this on save even if both values came from the existing instance
        # so there is an inefficiency here but it should be trivial
        try:
            validate_cert_with_key(cert_info['SP_PUBLIC_CERT'], cert_info['SP_PRIVATE_KEY'])
        except ValidationError as e:
            errors['SP_PRIVATE_KEY'] = e

        idp_data = attrs.get('ENABLED_IDPS', {}).get(idp_string, {})
        if not idp_data.get('attr_user_permanent_id', None) and not idp_data.get('attr_username'):
            errors['IDP_ATTR_USERNAME'] = "Either IDP_ATTR_USERNAME or IDP_ATTR_USER_PERMANENT_ID needs to be set"

        if errors:
            raise ValidationError(errors)

        saml_auth = SAMLAuth(AuthenticatorConfigTestStrategy(AuthenticatorStorage(), additional_settings=attrs))
        saml_auth.redirect_uri = attrs['CALLBACK_URL']
        idp = SAMLIdentityProvider(idp_string, **attrs['ENABLED_IDPS'][idp_string])
        config = saml_auth.generate_saml_config(idp=idp)
        invalid_security_settings = []
        try:
            settings = OneLogin_Saml2_Settings(settings=config)
            settings._security = {}
            settings._add_default_values()
            valid_security_settings = set(settings._security.keys())
            security_settings = set(attrs.get('SECURITY_CONFIG').keys())
            invalid_security_settings = security_settings.difference(valid_security_settings)
        except Exception as e:
            raise ValidationError(_("Failed to load config: %(e)s") % {"e": e})

        if invalid_security_settings:
            raise ValidationError({'SECURITY_CONFIG': _("Invalid keys:) {', '.join(invalid_security_settings)}")})

        response = super().validate(attrs)
        return response

    def to_internal_value(self, data):
        resp = super().to_internal_value(data)
        idp_data = {}
        for field, idp_field in self.settings_to_enabled_idps_fields.items():
            if field in resp:
                idp_data[idp_field] = resp[field]
                del resp[field]
        resp['ENABLED_IDPS'] = {idp_string: idp_data}
        return resp

    def to_representation(self, configuration):
        if 'ENABLED_IDPS' in configuration:
            for config_setting_name in self.settings_to_enabled_idps_fields:
                enabled_idp_field_name = self.settings_to_enabled_idps_fields[config_setting_name]
                if enabled_idp_field_name in configuration['ENABLED_IDPS'][idp_string]:
                    configuration[config_setting_name] = configuration['ENABLED_IDPS'][idp_string][enabled_idp_field_name]
            del configuration['ENABLED_IDPS']
        return configuration


class AuthenticatorPlugin(SocialAuthMixin, SocialAuthValidateCallbackMixin, SAMLAuth, AbstractAuthenticatorPlugin):
    configuration_class = SAMLConfiguration
    type = "SAML"
    logger = logger
    category = "sso"
    configuration_encrypted_fields = ['SP_PRIVATE_KEY']

    def generate_saml_config(self, idp=None):
        """
        SP Assertion Consumer Service URL (ACS URL) is where the IdP should send the login response (the URL back to the SP)

        tl; dr;
        This is a "super function" to fix setting the ACS URL

        The SP ACS URL is pulled form the config['sp']['assertionConsumerService']['url']
        This defaults to self.redirect_uri which, I believe, is made from inspecting the request.
        This became an issue when doing SSL offloading.
        Ideally, we want this to come from CALLBACK_URL which we expose as the ACS URL.
        This function overrides the default function setting the redirect_uri to our CALLBACK_URL
        """
        self.redirect_uri = self.strategy.get_setting('CALLBACK_URL', self)
        return super().generate_saml_config(idp=idp)

    def get_login_url(self, authenticator):
        url = get_relative_url('social:begin', kwargs={'backend': authenticator.slug})
        return f'{url}?idp={idp_string}'

    def _create_saml_auth(self, idp):
        """
        Get an instance of OneLogin_Saml2_Auth
        Identical to the implementation in SAMLAuth, with the exception
        of forcing https in the request info when SOCIAL_AUTH_REDIRECT_IS_HTTPS == True
        """
        config = self.generate_saml_config(idp)
        request_info = {
            "https": "on" if self.strategy.request_is_secure() else "off",
            "http_host": self.strategy.request_host(),
            "script_name": self.strategy.request_path(),
            "get_data": self.strategy.request_get(),
            "post_data": self.strategy.request_post(),
        }

        if self.strategy.setting("REDIRECT_IS_HTTPS", False, self):
            request_info["https"] = "on"

        return OneLogin_Saml2_Auth(request_info, config)

    def add_related_fields(self, request, authenticator):
        return {"metadata": get_relative_url('authenticator-metadata', kwargs={'pk': authenticator.id})}

    def extra_data(self, user, backend, response, *args, **kwargs):
        attrs = response["attributes"] if "attributes" in response else {}
        for perm in ["is_superuser", get_setting('ANSIBLE_BASE_SOCIAL_AUDITOR_FLAG')]:
            if perm in attrs:
                kwargs["social"].extra_data[perm] = attrs[perm]

        # Move group spec up a level if present
        if "Group" in attrs:
            response["Group"] = attrs["Group"]
        data = super().extra_data(user, backend, response, *args, **kwargs)

        # Ideally we would always have a DB instance
        # But if something was mocked in a test or somehow a db_instance just wasn't passed in we don't want to error here
        if self.database_instance is None:
            return data

        # The fields we want in extra get embedded into the ENABLED_IDPS field so we need to get them out from there
        idp_data = self.database_instance.configuration.get('ENABLED_IDPS', {}).get(idp_string, {})

        # We are going to auto-include all of the fields like USER_FIRST_NAME, USER_EMAIL, etc.
        for field, attr_name in SAMLConfiguration.settings_to_enabled_idps_fields.items():
            if field in ('IDP_URL', 'IDP_X509_CERT', 'IDP_ENTITY_ID'):
                continue

            field_name = idp_data.get(attr_name, None)
            if field_name in attrs:
                data[field_name] = attrs[field_name]

        return data

    def get_user_groups(self, extra_groups=[]):
        return extra_groups

    def get_alternative_uid(self, **kwargs):
        if uid := kwargs.get("uid", None):
            if ":" in uid:
                return uid.split(":", maxsplit=1)[1]

        return None


class SAMLMetadataView(View):
    def get(self, request, pk=None, format=None):
        authenticator = Authenticator.objects.get(id=pk)
        plugin = get_authenticator_plugin(authenticator.type)
        if plugin.type != 'SAML':
            logger.debug(f"Authenticator {authenticator.id} has a type which does not support metadata {plugin.type}")
            return HttpResponseNotFound()

        strategy = AuthenticatorStrategy(AuthenticatorStorage())
        complete_url = authenticator.configuration.get('CALLBACK_URL')
        saml_backend = strategy.get_backend(slug=authenticator.slug, redirect_uri=complete_url)
        try:
            metadata, errors = saml_backend.generate_metadata_xml()
        except OneLogin_Saml2_Error as e:
            errors = e
        if not errors:
            return HttpResponse(content=metadata, content_type='text/xml')
        else:
            return HttpResponse(content=errors, content_type='text/plain')


urls = [
    # SAML Metadata
    re_path(r'authenticators/(?P<pk>[0-9]+)/metadata/$', SAMLMetadataView.as_view(), name='authenticator-metadata'),
]



def apply_authentication_customizations() -> None:
    """Declare schema of DAB authentication classes

    This follows docs which reccomends OpenApiAuthenticationExtension to register an authentication class
    https://drf-spectacular.readthedocs.io/en/latest/customization.html#specify-authentication-with-openapiauthenticationextension
    As long as this class is resolved on import, drf-spectacular will be aware of it.
    This is called from api_documentation ready method.
    Imports are in-line, because dependencies may not be satisfied depending on what apps are installed.
    """
    from drf_spectacular.authentication import SessionScheme

    from ansible_base.authentication.session import SessionAuthentication

    class MyAuthenticationScheme(SessionScheme):
        target_class = SessionAuthentication
        name = 'SessionAuthentication'



import logging

from social_core.backends.github import GithubOrganizationOAuth2

from ansible_base.authentication.authenticator_configurators.github import GithubOrganizationConfiguration
from ansible_base.authentication.authenticator_plugins.base import AbstractAuthenticatorPlugin
from ansible_base.authentication.social_auth import SocialAuthMixin, SocialAuthValidateCallbackMixin

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.github_organization')


class AuthenticatorPlugin(SocialAuthMixin, SocialAuthValidateCallbackMixin, GithubOrganizationOAuth2, AbstractAuthenticatorPlugin):
    configuration_class = GithubOrganizationConfiguration
    logger = logger
    type = "github-org"
    category = "sso"
    configuration_encrypted_fields = ['SECRET']



# Generated by Django 4.2.16 on 2024-11-22 21:25

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dab_authentication', '0015_alter_authenticator_category_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authenticatoruser',
            name='access_allowed',
            field=models.BooleanField(default=None, help_text='Tracks if this user was allowed access to the system from the authenticator maps.', null=True),
        ),
        migrations.AlterField(
            model_name='authenticatoruser',
            name='claims',
            field=models.JSONField(blank=True, default=dict, help_text='The claims for the user as generated by the authenticator maps on their last login.'),
        ),
        migrations.AlterField(
            model_name='authenticatoruser',
            name='last_login_map_results',
            field=models.JSONField(blank=True, default=list, help_text='A data structure indicating how the authenticator maps were evaluated for the last login attempt.'),
        ),
        migrations.AlterField(
            model_name='authenticatoruser',
            name='provider',
            field=models.ForeignKey(help_text='The provider this user authenticated from.', on_delete=django.db.models.deletion.PROTECT, related_name='authenticator_providers', to='dab_authentication.authenticator', to_field='slug'),
        ),
        migrations.AlterField(
            model_name='authenticatoruser',
            name='user',
            field=models.ForeignKey(help_text='The local DB user related to this authenticator user.', on_delete=django.db.models.deletion.CASCADE, related_name='authenticator_users', to=settings.AUTH_USER_MODEL),
        ),
    ]



import hashlib

from ansible_base.lib.utils.hashing import hash_string


def hash_tokens(apps, schema_editor):
    OAuth2AccessToken = apps.get_model("dab_oauth2_provider", "OAuth2AccessToken")
    OAuth2RefreshToken = apps.get_model("dab_oauth2_provider", "OAuth2RefreshToken")
    for model in (OAuth2AccessToken, OAuth2RefreshToken):
        for token in model.objects.all():
            # Never re-hash a hashed token
            if token.token.startswith("$"):
                continue
            hashed = hash_string(token.token, hasher=hashlib.sha256, algo="sha256")
            token.token = hashed
            token.save()



# Generated by Django 4.2.8 on 2024-04-21 16:39

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dab_authentication', '0009_alter_authenticatoruser_provider_and_more'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='authenticatoruser',
            unique_together={('provider', 'uid')},
        ),
    ]



from cryptography.hazmat.primitives import serialization
from cryptography.x509 import load_pem_x509_certificate
from django.contrib.auth import get_user_model
from rest_framework import serializers

from ansible_base.lib.utils.encryption import ENCRYPTED_STRING
from ansible_base.lib.utils.validation import validate_url, validate_url_list


class UILabelMixIn:
    def __init__(self, **kwargs):
        self.ui_field_label = kwargs.pop('ui_field_label', 'Undefined')
        super().__init__(**kwargs)


class BooleanField(UILabelMixIn, serializers.BooleanField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class CharField(UILabelMixIn, serializers.CharField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class ChoiceField(UILabelMixIn, serializers.ChoiceField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class DictField(UILabelMixIn, serializers.DictField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class ListField(UILabelMixIn, serializers.ListField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class JSONField(UILabelMixIn, serializers.JSONField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class IntegerField(UILabelMixIn, serializers.IntegerField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class URLField(UILabelMixIn, serializers.CharField):
    def __init__(self, **kwargs):
        self.schemes = kwargs.pop('schemes', ['https', 'http'])
        self.allow_plain_hostname = kwargs.pop('allow_plain_hostname', True)
        super().__init__(**kwargs)

        def validator(value):
            return validate_url(value, schemes=self.schemes, allow_plain_hostname=self.allow_plain_hostname)

        self.validators.append(validator)


class URLListField(UILabelMixIn, serializers.ListField):
    def __init__(self, **kwargs):
        self.schemes = kwargs.pop('schemes', ['https', 'http'])
        self.allow_plain_hostname = kwargs.pop('allow_plain_hostname', True)
        super().__init__(**kwargs)

        def validator(value):
            return validate_url_list(value, schemes=self.schemes, allow_plain_hostname=self.allow_plain_hostname)

        self.validators.append(validator)


class UserAttrMap(UILabelMixIn, serializers.DictField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        def validator(value):
            errors = {}

            valid_user_attr_fields = set(["email", "username", "first_name", "last_name"])
            given_fields = set(list(value.keys()))

            missing_required_fields = set(get_user_model().REQUIRED_FIELDS) - given_fields
            for field in missing_required_fields:
                errors[field] = "Must be present"

            invalid_fields = given_fields - valid_user_attr_fields
            for field in invalid_fields:
                errors[field] = "Is not valid"

            valid_fields = given_fields.intersection(valid_user_attr_fields)
            for field in valid_fields:
                if type(value[field]) is not str:
                    errors[field] = "Must be a string"

            if errors:
                raise serializers.ValidationError(errors)

        self.validators.append(validator)


class PublicCert(UILabelMixIn, serializers.CharField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.public_cert = None

        def validator(value):
            if value:
                try:
                    self.public_cert = load_pem_x509_certificate(bytes(value, "UTF-8"))
                except Exception as e:
                    raise serializers.ValidationError(f"Unable to load as PEM data {e}")

        self.validators.append(validator)


class PrivateKey(UILabelMixIn, serializers.CharField):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.private_key = None

        def validator(value):
            if value and value != ENCRYPTED_STRING:
                try:
                    self.private_key = serialization.load_pem_private_key(bytes(value, "UTF-8"), password=None)
                except Exception as e:
                    raise serializers.ValidationError(f"Unable to load as PEM data {e}")

        self.validators.append(validator)



"""
Command to sync resources from RESOURCE_SERVER to local services
Each service must execute this command manually or scheduled.

Alternatively each service can use the functions implemented on
resource_registry.tasks module to spawn syncs in a task queue.

Usage::

    django-admin resource_sync  # sync all resources

Optional resource type name argument::

    django-admin resource_sync user  # only shared.user
    django-admin resource_sync shared.user  # the same
    django-admin resource_sync user organization team  # shared. prefix will be handled

Optional parameters::

    `--retries number` to retry failed syncs
    `--retrysleep seconds` to set interval between retries
    `--asyncio` Flag to enable asyncio executor
"""

from django.core.management.base import BaseCommand, CommandError

from ansible_base.resource_registry.tasks.sync import ResourceSyncHTTPError, SyncExecutor, get_resource_type_names


def valid_resource_type(resource_type):  # pragma: no cover
    resource_type = resource_type.strip()
    if not resource_type.startswith("shared."):
        resource_type = f"shared.{resource_type}"
    valid_options = get_resource_type_names()
    if resource_type not in valid_options:
        raise CommandError(f"Invalid resource_type {resource_type}, options are {valid_options}")
    return resource_type


class Command(BaseCommand):  # pragma: no cover
    help = "Fetch Resource State from RESOURCE_PROVIDER and sync with local service."

    def add_arguments(self, parser):
        parser.add_argument(
            "resource_type_names",
            nargs="*",
            type=valid_resource_type,
            help="Optional, one or more names e.g: `shared.user` or `shared.user shared.organization`",
        )
        parser.add_argument(
            "--retries",
            type=int,
            default=3,
            help="For when a resource is unavailable, how many retries to perform",
            required=False,
        )
        parser.add_argument(
            "--retrysleep",
            type=int,
            default=30,
            help="Interval between retries",
            required=False,
        )
        parser.add_argument("--asyncio", action="store_true", default=False, help="Enable asyncio executor")

    def handle(self, *args, **options):
        """Handle RESOURCE_PROVIDER sync"""
        arguments = ["resource_type_names", "retries", "retrysleep", "asyncio"]
        options = {k: v for k, v in options.items() if k in arguments}
        try:
            executor = SyncExecutor(**options, stdout=self.stdout)
            executor.run()
        except ResourceSyncHTTPError as exc:
            raise CommandError(f"Error accessing Resource Server: {str(exc)}")

        # NOTE: Can optionally take --report-json and then serialize executor.results



from django.apps import apps
from django.conf import settings
from django.db.models import Model
from django.db.models.query import QuerySet
from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import PermissionDenied

from ansible_base.lib.utils.settings import get_setting
from ansible_base.rbac.evaluations import has_super_permission
from ansible_base.rbac.models import ObjectRole
from ansible_base.rbac.permission_registry import permission_registry
from ansible_base.rbac.validators import permissions_allowed_for_role


def visible_users(request_user, queryset=None, always_show_superusers=True, always_show_self=True) -> QuerySet:
    """Gives a queryset of users that another user should be able to view"""
    user_cls = permission_registry.user_model
    org_cls = apps.get_model(settings.ANSIBLE_BASE_ORGANIZATION_MODEL)

    if can_view_all_users(request_user):
        if queryset is not None:
            return queryset
        else:
            return user_cls.objects.all()

    object_id_fd = ObjectRole._meta.get_field('object_id')
    members_of_visible_orgs = ObjectRole.objects.filter(
        role_definition__permissions__codename='member_organization', object_id__in=org_cls.access_ids_qs(request_user, 'view', cast_field=object_id_fd)
    ).values('users')
    if queryset is None:
        queryset = user_cls.objects

    queryset = queryset.filter(pk__in=members_of_visible_orgs)
    if always_show_superusers:
        queryset = queryset | user_cls.objects.filter(is_superuser=True)
    if always_show_self:
        queryset = queryset | user_cls.objects.filter(pk=request_user.id)
    return queryset.distinct()


def can_view_all_users(request_user):
    org_cls = apps.get_model(settings.ANSIBLE_BASE_ORGANIZATION_MODEL)

    return has_super_permission(request_user, 'view') or (
        get_setting('ORG_ADMINS_CAN_SEE_ALL_USERS', False) and org_cls.access_ids_qs(request_user, 'change').exists()
    )


def can_change_user(request_user, target_user) -> bool:
    """Tells if the request user can modify details of the target user"""
    if request_user.is_superuser:
        return True
    elif target_user.is_superuser:
        return False  # target is a superuser and request user is not

    if not get_setting('MANAGE_ORGANIZATION_AUTH', False):
        return False

    # All users can change their own password and other details
    if request_user.pk == target_user.pk:
        return True

    # If the user is not in any organizations, answer can not consider organization permissions
    org_cls = apps.get_model(settings.ANSIBLE_BASE_ORGANIZATION_MODEL)
    target_user_orgs = org_cls.access_qs(target_user, 'member_organization')
    if not target_user_orgs.exists():
        return request_user.is_superuser

    # Organization admins can manage users in their organization
    # this requires change permission to all organizations the target user is a member of
    return not target_user_orgs.exclude(pk__in=org_cls.access_ids_qs(request_user, 'change_organization')).exists()


def check_content_obj_permission(request_user, obj) -> None:
    """Permission policy rules for giving or removing obj permission

    Right now we are not supporting a separate permission to manage permission
    on objects, so we firstly look to a simple matter of having change permission
    If that is not available, then we check all object-level permissions.
    """
    if 'change' in obj._meta.default_permissions:
        # Model has no change permission, so user must have all permissions for the applicable model
        if not request_user.has_obj_perm(obj, 'change'):
            raise PermissionDenied
    else:
        cls = type(obj)
        for codename in permissions_allowed_for_role(cls)[cls]:
            if not request_user.has_obj_perm(obj, codename):
                raise PermissionDenied({'detail': _('You do not have {codename} permission the object').format(codename=codename)})


def check_can_remove_assignment(request_user: Model, assignment: Model):
    """Removing a role assignment will OR checks for the actor and the object

    You can remove a permission if you can manage the user or team given the role
    OR, if you have change permission to the content object targeted by the assignment.
    """
    if request_user.is_superuser:
        return

    assignment_model_name = assignment._meta.model_name
    if assignment_model_name == 'roleuserassignment':
        if can_change_user(request_user, assignment.user):
            return
    elif assignment_model_name == 'roleteamassignment':
        if request_user.has_obj_perm(assignment.team, 'change'):
            return
    else:
        raise RuntimeError(f'Assignment model {assignment_model_name} not recognized as a role assignment model')

    # request user is not a manager of the actor of the assignment
    # but can still remove the assignment if they manage the content object it applies to
    if assignment.content_type_id:
        check_content_obj_permission(request_user, assignment.content_object)
    else:
        # Case of a system role with a non-superuser user
        raise PermissionDenied



from django.apps import AppConfig

from ansible_base.rbac.permission_registry import permission_registry


class AnsibleRBACConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'ansible_base.rbac'
    label = 'dab_rbac'
    verbose_name = 'DAB shared RBAC'

    def ready(self):
        permission_registry.call_when_apps_ready(self.apps, self)



from django.apps import AppConfig


class HelpTextCheckConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'ansible_base.help_text_check'
    label = 'dab_help_text_check'
    verbose_name = 'Django Model Help Text Checker'



from rest_framework import serializers

from ansible_base.resource_registry.utils.resource_type_serializers import AnsibleResourceForeignKeyField, SharedResourceTypeSerializer
from ansible_base.resource_registry.utils.sso_provider import get_sso_provider_server


class UserAdditionalDataSerializer(serializers.Serializer):
    """
    Additional data serializer for UserType
    """

    social_auth = serializers.ListField()

    def to_representation(self, instance):
        social_auth = []

        if hasattr(instance, "authenticator_users"):
            for social in instance.authenticator_users.all():
                sso_server, uid = get_sso_provider_server(social.provider.slug, social.uid)
                social_auth.append(
                    {
                        "uid": uid,
                        "backend_type": social.provider.type,
                        "sso_server": sso_server,
                    }
                )
        elif hasattr(instance, "social_auth"):
            for social in instance.social_auth.all():
                sso_server, uid = get_sso_provider_server(social.provider, social.uid)
                social_auth.append(
                    {
                        "uid": uid,
                        "backend_type": social.provider,
                        "sso_server": sso_server,
                    }
                )

        return {"social_auth": social_auth}


class UserType(SharedResourceTypeSerializer):
    RESOURCE_TYPE = "user"
    ADDITIONAL_DATA_SERIALIZER = UserAdditionalDataSerializer
    UNIQUE_FIELDS = ("username",)

    username = serializers.CharField()
    email = serializers.EmailField(required=False, allow_blank=True)
    first_name = serializers.CharField(required=False, allow_blank=True)
    last_name = serializers.CharField(required=False, allow_blank=True)
    # Commenting this out for now because Galaxy NG doesn't have a system auditor flag
    # is_platform_auditor = serializers.BooleanField()


class OrganizationType(SharedResourceTypeSerializer):
    RESOURCE_TYPE = "organization"
    UNIQUE_FIELDS = ("name",)

    name = serializers.CharField()
    description = serializers.CharField(
        default="",
        allow_blank=True,
    )


class TeamType(SharedResourceTypeSerializer):
    RESOURCE_TYPE = "team"
    UNIQUE_FIELDS = (
        "name",
        "organization",
    )

    name = serializers.CharField()
    organization = AnsibleResourceForeignKeyField("shared.organization", required=False, allow_null=True)
    description = serializers.CharField(
        default="",
        allow_blank=True,
    )



# Generated by Django 4.2.16 on 2024-11-21 11:14

import ansible_base.resource_registry.models.service_identifier
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dab_resource_registry', '0005_resource_is_partially_migrated_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='resource',
            name='service_id',
            field=models.UUIDField(default=ansible_base.resource_registry.models.service_identifier.service_id, help_text='The ID of the service responsible for managing this resource.'),
        ),
    ]
